import asyncio
import collections
import errno
import os
import re
from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import errors
from upsies.jobs import JobBase
from upsies.utils import signal


class FooJob(JobBase):
    name = 'foo'
    label = 'Foo'

    def initialize(self, foo=None, bar=None):
        assert self.name == 'foo'
        assert self.label == 'Foo'
        assert self.output == ()
        self.initialize_was_called = True

    def execute(self):
        assert self.name == 'foo'
        assert self.label == 'Foo'
        assert self.output == ()
        self.execute_was_called = True


@pytest.fixture
def job(tmp_path):
    return FooJob(home_directory=tmp_path, cache_directory=tmp_path, ignore_cache=False)


@pytest.mark.parametrize(
    argnames='path, exp_exception',
    argvalues=(
        ('path/to/home', None),
        ('', None),
        (None, None),
        ('/root/upsies/test', errors.ContentError('/root/upsies/test: Permission denied')),
    ),
)
def test_home_directory_property(path, exp_exception, tmp_path):
    if path is None:
        job = FooJob()
        assert job.home_directory == ''
    else:
        job = FooJob(home_directory=tmp_path / path)
        if exp_exception:
            with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
                job.home_directory
        else:
            assert job.home_directory == tmp_path / path
            assert os.path.exists(job.home_directory)


@pytest.mark.parametrize(
    argnames='path',
    argvalues=(
        'path/to/cache',
        '',
        None,
    ),
)
def test_cache_directory_property(path, tmp_path, mocker):
    mocker.patch('upsies.constants.DEFAULT_CACHE_DIRECTORY', tmp_path / 'cache/path')
    if not path:
        job = FooJob()
        assert job.cache_directory == tmp_path / 'cache/path'
    else:
        job = FooJob(cache_directory=tmp_path / path)
        assert job.cache_directory == tmp_path / path
        assert os.path.exists(job.cache_directory)


def test_ignore_cache_property(tmp_path):
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, ignore_cache=False).ignore_cache is False
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, ignore_cache=True).ignore_cache is True
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, ignore_cache='').ignore_cache is False
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, ignore_cache=1).ignore_cache is True


def test_no_output_is_ok_property(tmp_path):
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, no_output_is_ok=False).no_output_is_ok is False
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, no_output_is_ok=True).no_output_is_ok is True
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, no_output_is_ok=0).no_output_is_ok is False
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, no_output_is_ok='maybe').no_output_is_ok is True


def test_hidden_property(tmp_path):
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, hidden=False).hidden is False
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, hidden=True).hidden is True
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, hidden='').hidden is False
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, hidden=1).hidden is True


@pytest.mark.parametrize(
    argnames='autostart_argument, exp_autostart',
    argvalues=(
        (True, True),
        (1, True),
        (False, False),
        ('', False),
    ),
)
def test_autostart(autostart_argument, exp_autostart, tmp_path):
    job = FooJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        autostart=autostart_argument,
    )
    assert job.autostart is exp_autostart


precondition_test_cases = pytest.mark.parametrize(
    argnames='precondition, exp_exception',
    argvalues=(
        (Mock(), None),
        ('foo', TypeError("Not callable: 'foo'")),
        ('', TypeError("Not callable: ''")),
        (True, TypeError("Not callable: True")),
        (False, TypeError("Not callable: False")),
    ),
)

@precondition_test_cases
def test_precondition_argument(precondition, exp_exception, tmp_path):
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            FooJob(home_directory=tmp_path, cache_directory=tmp_path, precondition=precondition)
    else:
        job = FooJob(home_directory=tmp_path, cache_directory=tmp_path, precondition=precondition)
        assert job.precondition is precondition


@pytest.mark.parametrize(
    argnames='is_started, exp_RuntimeError',
    argvalues=(
        (False, None),
        (True, RuntimeError('Cannot set precondition after job has been started')),
    ),
)
@precondition_test_cases
def test_precondition_property(precondition, exp_exception, is_started, exp_RuntimeError, tmp_path, mocker):
    job = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    mocker.patch.object(type(job), 'is_started', PropertyMock(return_value=is_started))
    if exp_RuntimeError:
        with pytest.raises(type(exp_RuntimeError), match=rf'^{re.escape(str(exp_RuntimeError))}$'):
            job.precondition = precondition
    elif exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            job.precondition = precondition
    else:
        job.precondition = precondition
        assert job.precondition is precondition


@pytest.mark.parametrize(
    argnames='precondition, prejobs, exp_is_enabled',
    argvalues=(
        (Mock(return_value=True), (Mock(is_finished=True, is_enabled=True), Mock(is_finished=True, is_enabled=True)), True),
        (Mock(return_value=False), (Mock(is_finished=True, is_enabled=True), Mock(is_finished=True, is_enabled=True)), False),

        (Mock(return_value=True), (Mock(is_finished=True, is_enabled=True), Mock(is_finished=True, is_enabled=False)), True),
        (Mock(return_value=True), (Mock(is_finished=True, is_enabled=True), Mock(is_finished=False, is_enabled=True)), False),
        (Mock(return_value=True), (Mock(is_finished=True, is_enabled=True), Mock(is_finished=False, is_enabled=False)), True),

        (Mock(return_value=True), (Mock(is_finished=True, is_enabled=False), Mock(is_finished=True, is_enabled=True)), True),
        (Mock(return_value=True), (Mock(is_finished=False, is_enabled=True), Mock(is_finished=True, is_enabled=True)), False),
        (Mock(return_value=True), (Mock(is_finished=False, is_enabled=False), Mock(is_finished=True, is_enabled=True)), True),
    ),
)
def test_is_enabled_property(precondition, prejobs, exp_is_enabled, tmp_path):
    job = FooJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        precondition=precondition,
        prejobs=prejobs,
    )
    assert job.is_enabled is exp_is_enabled


def test_kwargs_property(tmp_path):
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, foo='a').kwargs.get('foo') == 'a'
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, foo='a').kwargs.get('bar') is None
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, bar='b').kwargs.get('bar') == 'b'
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, bar='b').kwargs.get('foo') is None
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, foo='a', bar='b').kwargs.get('foo') == 'a'
    assert FooJob(home_directory=tmp_path, cache_directory=tmp_path, foo='a', bar='b').kwargs.get('bar') == 'b'


def test_signal_property(job):
    assert isinstance(job.signal, signal.Signal)
    assert set(job.signal.signals) == {
        'output',
        'info',
        'warning',
        'error',
        'executed',
        'finished',
        'refresh_job_list',
    }
    assert set(job.signal.recording) == {'output'}


@pytest.mark.parametrize(
    argnames='is_started, exp_RuntimeError',
    argvalues=(
        (False, None),
        (True, RuntimeError('Cannot set prejobs after job has been started')),
    ),
)
def test_prejobs_property(is_started, exp_RuntimeError, job, mocker):
    assert isinstance(job.prejobs, tuple)
    assert job.prejobs is job._prejobs

    mocker.patch.object(type(job), 'is_started', PropertyMock(return_value=is_started))
    if exp_RuntimeError:
        with pytest.raises(type(exp_RuntimeError), match=rf'^{re.escape(str(exp_RuntimeError))}$'):
            job.prejobs = [1, 2, 3]
    else:
        job.prejobs = [1, 2, 3]
        assert job.prejobs == (1, 2, 3)


@pytest.mark.parametrize(
    argnames='presignals, emissions, exp_is_enabled_after_all_emissions',
    argvalues=(
        (
            [],
            [],
            True,
        ),
        (
            [],
            [
                {'emission': ('info', 'Info.'), 'exp_is_enabled': True, 'refresh_job_list_calls': []},
            ],
            True,
        ),
        (
            ['info'],
            [
                {'emission': ('output', 'Output.'), 'exp_is_enabled': False, 'refresh_job_list_calls': []},
                {'emission': ('info', 'Info.'), 'exp_is_enabled': True, 'refresh_job_list_calls': [call()]},
                {'emission': ('warning', 'Warning.'), 'exp_is_enabled': True, 'refresh_job_list_calls': []},
            ],
            True,
        ),
        (
            ['info', 'output'],
            [
                {'emission': ('output', 'Output 1.'), 'exp_is_enabled': False, 'refresh_job_list_calls': []},
                {'emission': ('warning', 'Warning 1.'), 'exp_is_enabled': False, 'refresh_job_list_calls': []},
                {'emission': ('output', 'Output 2.'), 'exp_is_enabled': False, 'refresh_job_list_calls': []},
                {'emission': ('info', 'Info 1.'), 'exp_is_enabled': True, 'refresh_job_list_calls': [call()]},
                {'emission': ('warning', 'Warning 2.'), 'exp_is_enabled': True, 'refresh_job_list_calls': []},
            ],
            True,
        ),
        (
            ['info', 'output'],
            [
                {'emission': ('output', 'Output 1.'), 'exp_is_enabled': False, 'refresh_job_list_calls': []},
                {'emission': ('warning', 'Warning 1.'), 'exp_is_enabled': False, 'refresh_job_list_calls': []},
                {'emission': ('output', 'Output 2.'), 'exp_is_enabled': False, 'refresh_job_list_calls': []},
                {'emission': ('warning', 'Warning 2.'), 'exp_is_enabled': False, 'refresh_job_list_calls': []},
            ],
            False,
        ),
    ),
)
def test_presignal(presignals, emissions, exp_is_enabled_after_all_emissions, job):
    job_list_refreshed = Mock()
    job.signal.register('refresh_job_list', job_list_refreshed)

    class BarJob(JobBase):
        name = 'bar'
        label = 'Bar'

    other_job = BarJob()

    for presignal in presignals:
        job.presignal(other_job, presignal)

    # Check registered presignals
    if presignals:
        assert dict(job.presignals) == {
            other_job: {ps: False for ps in presignals}
        }
    else:
        assert dict(job.presignals) == {}

    assert job.is_enabled is (False if presignals else True)

    # Emit emissions in random order
    for state in emissions:
        signal, payload = state['emission']
        other_job.signal.emit(signal, payload)

        assert job.is_enabled is state['exp_is_enabled']

        assert job_list_refreshed.call_args_list == state['refresh_job_list_calls']
        job_list_refreshed.reset_mock()

    assert job.is_enabled is exp_is_enabled_after_all_emissions

    # Check registered presignals
    emitted_signals = list(set(
        state['emission'][0]
        for state in emissions
        if state['emission'][0] in presignals
    ))
    if presignals:
        exp_presignals = {
            other_job: {ps: ps in emitted_signals for ps in presignals}
        }
    else:
        exp_presignals = {}
    assert dict(job.presignals) == dict(exp_presignals)

@pytest.mark.parametrize(
    argnames='is_started, exp_RuntimeError',
    argvalues=(
        (False, None),
        (True, RuntimeError('Cannot set presignal after job has been started')),
    ),
)
def test_presignal_after_job_is_started(is_started, exp_RuntimeError, job, mocker):
    class BarJob(JobBase):
        name = 'bar'
        label = 'Bar'

    other_job = BarJob()

    mocker.patch.object(type(job), 'is_started', PropertyMock(return_value=is_started))
    if exp_RuntimeError:
        with pytest.raises(type(exp_RuntimeError), match=rf'^{re.escape(str(exp_RuntimeError))}$'):
            job.presignal(other_job, 'finished')
    else:
        job.presignal(other_job, 'finished')


def test_initialize_is_called_after_object_creation(job):
    assert job.initialize_was_called


def test_callbacks_argument(tmp_path):
    class BarJob(FooJob):
        def initialize(self):
            self.signal.add('greeted')

    cb = Mock()
    job = BarJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        callbacks={
            'output': cb.output,
            'finished': cb.finished,
            'greeted': (cb.hello, cb.hey),
        },
    )
    assert cb.output in job.signal.signals['output']
    assert cb.finished in job.signal.signals['finished']
    assert cb.hello in job.signal.signals['greeted']
    assert cb.hey in job.signal.signals['greeted']


def test_presignals_argument(mocker):
    class BarJob(FooJob):
        def initialize(self):
            self.signal.add('a')
            self.signal.add('b')
            self.signal.add('c')

    class BazJob(FooJob):
        def initialize(self):
            self.signal.add('baz')

    mocker.patch('upsies.jobs.base.JobBase.presignal')

    presignals = {
        BarJob(): ['a', 'c'],
        BazJob(): ['baz'],
    }
    job = BarJob(presignals=presignals)
    assert job.presignal.call_args_list == [
        call(job, signal)
        for job, signals in presignals.items()
        for signal in signals
    ]


def test_start_does_nothing_if_job_is_already_finished(job):
    job.finish()
    assert job.is_finished is True
    job.start()
    assert job.is_started is False
    assert job.is_finished is True

def test_start_does_nothing_if_job_is_not_enabled(job):
    job.precondition = lambda: False
    assert job.is_enabled is False
    assert job.is_started is False
    job.start()
    assert job.is_started is False

def test_start_sets_started_property(job):
    assert job.is_started is False
    job.start()
    assert job.is_started is True

def test_start_is_called_multiple_times(job):
    job.start()
    with pytest.raises(RuntimeError, match=r'^start\(\) was already called$'):
        job.start()

def test_start_reads_cache_from_previous_job(tmp_path):
    class BarJob(FooJob):
        def _read_cache(self):
            return True

    executed_cb = Mock()
    job = BarJob(home_directory=tmp_path, cache_directory=tmp_path, callbacks={'executed': executed_cb})
    assert executed_cb.call_args_list == []
    job.start()
    assert executed_cb.call_args_list == []
    assert not hasattr(job, 'execute_was_called')
    assert job.is_finished is True

def test_start_finishes_if_reading_from_cache_fails(tmp_path):
    class BarJob(FooJob):
        def _read_cache(self):
            raise RuntimeError('I found god')

    job = BarJob(home_directory=tmp_path, cache_directory=tmp_path)
    with pytest.raises(RuntimeError, match=r'I found god'):
        job.start()
    assert job.is_finished is True
    assert not hasattr(job, 'execute_was_called')

def test_start_executes_job_if_no_cached_output_is_found(job):
    executed_cb = Mock()
    job.signal.register('executed', executed_cb)
    assert executed_cb.call_args_list == []
    job.start()
    assert executed_cb.call_args_list == [call(job)]
    assert job.execute_was_called is True
    assert job.is_finished is False


@pytest.mark.asyncio
async def test_wait_returns_when_finish_is_called(tmp_path):
    class BarJob(FooJob):
        calls = collections.deque()

        def execute(self):
            asyncio.get_running_loop().call_later(0.1, self.finish)

        def finish(self):
            self.calls.append('finish() called')
            super().finish()

        async def wait(self):
            await super().wait()
            self.calls.append('wait() returned')

    job = BarJob(home_directory=tmp_path, cache_directory=tmp_path)
    job.start()
    await job.wait()
    assert list(job.calls) == ['finish() called', 'wait() returned']


@pytest.mark.parametrize(
    argnames='is_started, is_finished, is_enabled, exp_finished',
    argvalues=(
        (True, True, True, False),
        (True, True, False, False),
        (True, False, True, True),
        (False, True, True, False),
        (True, False, False, True),
        (False, False, True, True),
        (False, True, False, False),
        (False, False, False, True),
    ),
)
@pytest.mark.asyncio
async def test_finish(is_started, is_finished, is_enabled, exp_finished, job, mocker):
    mocker.patch.object(type(job), 'is_started', PropertyMock(return_value=is_started))
    mocker.patch.object(type(job), 'is_finished', PropertyMock(return_value=is_finished))
    mocker.patch.object(type(job), 'is_enabled', PropertyMock(return_value=is_enabled))

    cb = Mock()
    job.signal.register('finished', cb)
    mocker.patch.object(job, '_write_cache')
    job._tasks = [
        Mock(done=Mock(return_value=False)),
        Mock(done=Mock(return_value=True)),
        Mock(done=Mock(return_value=False)),
    ]

    job.start()

    for i in range(1):
        job.finish()
        if exp_finished:
            assert cb.call_args_list == [call(job)]

            # for task in job._tasks:
            #     assert task.done.call_args_list == [call()]
            # assert job._tasks[0].cancel.call_args_list == [call()]
            # assert job._tasks[1].cancel.call_args_list == []
            # assert job._tasks[2].cancel.call_args_list == [call()]

            assert job._write_cache.call_args_list == [call()]
        else:
            assert cb.call_args_list == []

            # for task in job._tasks:
            #     assert task.done.call_args_list == []
            # assert job._tasks[0].cancel.call_args_list == []
            # assert job._tasks[1].cancel.call_args_list == []
            # assert job._tasks[2].cancel.call_args_list == []

            assert job._write_cache.call_args_list == []


@pytest.mark.asyncio
async def test_is_finished(job):
    asyncio.get_running_loop().call_soon(job.finish)
    job.start()
    assert job.is_finished is False
    await job.wait()
    assert job.is_finished is True


@pytest.mark.parametrize('is_finished', (True, False))
@pytest.mark.parametrize(
    argnames='errors, raised, output, no_output_is_ok, exp_exit_code',
    argvalues=(
        ([Exception('Something went wrong')], [], ['Some output'], False, 1),
        ([], [Exception('Something went really bad')], ['Some output'], False, 1),
        ([], [], ['Some output'], False, 0),
        ([], [], [], False, 1),
        ([], [], [], True, 0),
    ),
    ids=lambda v: str(v),
)
def test_exit_code_is_None_if_job_is_not_finished(is_finished, errors, raised, output, no_output_is_ok,
                                                  exp_exit_code,
                                                  job, mocker):
    mocker.patch.object(type(job), 'is_finished', PropertyMock(return_value=is_finished))
    mocker.patch.object(type(job), 'errors', PropertyMock(return_value=errors))
    mocker.patch.object(type(job), 'raised', PropertyMock(return_value=raised))
    mocker.patch.object(type(job), 'output', PropertyMock(return_value=output))
    mocker.patch.object(type(job), 'no_output_is_ok', PropertyMock(return_value=no_output_is_ok))

    if not is_finished:
        assert job.exit_code is None
    else:
        assert job.exit_code is exp_exit_code


def test_send_emits_output_signal(job, mocker):
    mocker.patch.object(job.signal, 'emit')
    job.start()
    assert job.signal.emit.call_args_list == [
        call('executed', job),
    ]
    job.send('foo')
    assert job.signal.emit.call_args_list == [
        call('executed', job),
        call('output', 'foo'),
    ]
    job.send([1, 2])
    assert job.signal.emit.call_args_list == [
        call('executed', job),
        call('output', 'foo'),
        call('output', '[1, 2]'),
    ]
    job.send('')
    assert job.signal.emit.call_args_list == [
        call('executed', job),
        call('output', 'foo'),
        call('output', '[1, 2]'),
        call('output', ''),
    ]

def test_send_on_finished_job(job):
    job.start()
    job.send('foo')
    job.finish()
    assert job.output == ('foo',)
    job.send('bar')
    assert job.output == ('foo',)

def test_send_fills_output_property(job, mocker):
    job.start()
    assert job.output == ()
    job.send('foo')
    assert job.output == ('foo',)
    job.send([1, 2])
    assert job.output == ('foo', '[1, 2]')


def test_output(job):
    job.start()
    assert job.output == ()
    job._output = ['foo', 'bar', 'baz']
    assert job.output == ('foo', 'bar', 'baz')


def test_info(job):
    job.start()
    cb = Mock()
    job.signal.register('info', cb)
    assert job.info == ''
    assert cb.call_args_list == []
    job.info = 'foo'
    assert job.info == 'foo'
    assert cb.call_args_list == [call('foo')]


def test_warn(job):
    job.start()
    assert job.warnings == ()
    job.warn('foo')
    assert job.warnings == ('foo',)
    job.warn('bar')
    assert job.warnings == ('foo', 'bar')
    job.finish()
    job.warn('baz')
    assert job.warnings == ('foo', 'bar')

def test_warnings(job):
    job.start()
    assert job.warnings == ()
    job._warnings = ['a', 'b', 'c']
    assert job.warnings == ('a', 'b', 'c')

def test_clear_warnings_on_unfinished_job(job):
    job.start()
    job.clear_warnings()
    assert job.warnings == ()
    job.warn('foo')
    assert job.warnings == ('foo',)
    job.clear_warnings()
    assert job.warnings == ()

def test_clear_warnings_on_finished_job(job):
    job.start()
    job.warn('foo')
    assert job.warnings == ('foo',)
    job.finish()
    job.clear_warnings()
    assert job.warnings == ('foo',)


def test_errors(job):
    job.start()
    assert job.errors == ()
    job._errors = ['foo', 'bar', 'baz']
    assert job.errors == ('foo', 'bar', 'baz')

@pytest.mark.parametrize('error', ('foo', errors.ContentError('bar')))
def test_error_emits_error_signal(error, job, mocker):
    mocker.patch.object(job.signal, 'emit')
    job.start()
    assert job.signal.emit.call_args_list == [
        call('executed', job),
    ]
    job.error(error)
    assert job.signal.emit.call_args_list == [
        call('executed', job),
        call('error', error),
        call('finished', job),
    ]

@pytest.mark.parametrize(
    argnames='finish, exp_finished',
    argvalues=(
        (None, True),
        (True, True),
        (False, False),
    ),
)
def test_error_finishes_job(finish, exp_finished, job, mocker):
    job.start()
    assert not job.is_finished
    if finish is None:
        job.error('foo')
    else:
        job.error('foo', finish=finish)
    assert job.is_finished is exp_finished

def test_error_on_finished_job(job, mocker):
    mocker.patch.object(job.signal, 'emit')
    job.start()
    assert job.signal.emit.call_args_list == [
        call('executed', job),
    ]
    job.finish()
    assert job.signal.emit.call_args_list == [
        call('executed', job),
        call('finished', job),
    ]
    job.error('foo')
    assert job.signal.emit.call_args_list == [
        call('executed', job),
        call('finished', job),
    ]

@pytest.mark.parametrize('error', ('foo', errors.ContentError('bar')))
def test_error_appends_to_errors_property(error, job, mocker):
    job.start()
    assert job.errors == ()
    job.error(error)
    assert job.errors == (error,)


@pytest.mark.asyncio
async def test_exception_on_finished_job(job):
    job.start()
    job.finish()
    job.exception(TypeError('Sorry, not my type.'))
    assert job.is_finished
    for _ in range(5):
        with pytest.raises(TypeError, match=r'^Sorry, not my type\.$'):
            await job.wait()  # No exception
        assert type(job.raised) is TypeError
        assert str(job.raised) == 'Sorry, not my type.'

@pytest.mark.asyncio
async def test_exception_sets_exception_that_is_raised_by_wait(job):
    job.start()
    job.exception(TypeError('Sorry, not my type.'))
    for _ in range(5):
        with pytest.raises(TypeError, match=r'^Sorry, not my type\.$'):
            await job.wait()
        assert type(job.raised) is TypeError
        assert str(job.raised) == 'Sorry, not my type.'

@pytest.mark.asyncio
async def test_exception_finishes_job(job):
    job.start()
    assert not job.is_finished
    job.exception(TypeError('Sorry, not my type.'))
    assert job.is_finished


@pytest.mark.asyncio
async def test_attach_task_to_finished_job(job, mocker):
    job.start()
    job.finish()
    with pytest.raises(RuntimeError, match=rf'^Job is already finished: {job.name}$'):
        job.attach_task('mock coro')
    assert job._tasks == []

@pytest.mark.asyncio
async def test_attached_task_returns_return_value(job, mocker):
    job.start()
    coro = AsyncMock(return_value='foo')()
    job.attach_task(coro)
    assert await job._tasks[0] == 'foo'
    assert not job.is_finished
    assert job.raised is None

@pytest.mark.asyncio
async def test_attached_task_passes_return_value_to_callback(job, mocker):
    job.start()
    callback = Mock()
    coro = AsyncMock(return_value='foo')()
    job.attach_task(coro, callback=callback)
    assert await job._tasks[0] == 'foo'
    assert not job.is_finished
    assert job.raised is None
    assert callback.call_args_list == [call('foo')]

@pytest.mark.asyncio
async def test_attached_task_catches_exceptions_from_coro(job, mocker):
    job.start()
    callback = Mock()
    coro = AsyncMock(side_effect=TypeError('foo'))()
    job.attach_task(coro, callback=callback)
    with pytest.raises(TypeError, match=r'^foo$'):
        await job._tasks[0]
    assert job.is_finished
    assert isinstance(job.raised, TypeError)
    assert str(job.raised) == 'foo'
    assert callback.call_args_list == []

@pytest.mark.asyncio
async def test_attached_task_catches_exceptions_from_callback(job, mocker):
    job.start()
    callback = Mock(side_effect=TypeError('foo'))
    coro = AsyncMock(return_value='foo')()
    job.attach_task(coro, callback=callback)
    await job._tasks[0]
    assert job.is_finished
    assert isinstance(job.raised, TypeError)
    assert str(job.raised) == 'foo'
    assert callback.call_args_list == [call('foo')]

@pytest.mark.parametrize('finish_when_done', (False, True))
@pytest.mark.asyncio
async def test_attached_task_ignores_CancelledError(finish_when_done, job, mocker):
    job.start()
    callback = Mock()
    coro = AsyncMock(side_effect=asyncio.CancelledError())()
    job.attach_task(coro, callback=callback, finish_when_done=finish_when_done)
    await job._tasks[0]  # No CancelledError raised
    assert job.is_finished == finish_when_done
    assert job.raised is None
    assert callback.call_args_list == [call(None)]

@pytest.mark.asyncio
async def test_await_tasks(job, mocker):
    job._tasks = [
        asyncio.create_task(AsyncMock()())
        for i in range(3)
    ]
    for task in job._tasks:
        assert not task.done()
    await job.await_tasks()
    for task in job._tasks:
        assert task.done()


@pytest.mark.parametrize(
    argnames='emissions, exit_code, cache_file, is_executed',
    argvalues=(
        ((), 0, 'path/to/cache_file', True),
        ((('output', {'args': ('foo',), 'kwargs': {'bar': 'baz'}}),), 1, 'path/to/cache_file', True),
        ((('output', {'args': ('foo',), 'kwargs': {'bar': 'baz'}}),), 0, '', True),
        ((('output', {'args': ('foo',), 'kwargs': {'bar': 'baz'}}),), 0, None, True),
        ((('output', {'args': ('foo',), 'kwargs': {'bar': 'baz'}}),), 1, 'path/to/cache_file', False),
    ),
)
def test_write_cache_does_nothing(emissions, exit_code, cache_file, is_executed, job, mocker):
    mocker.patch.object(type(job.signal), 'emissions', PropertyMock(return_value=emissions))
    mocker.patch.object(type(job), 'exit_code', PropertyMock(return_value=exit_code))
    mocker.patch.object(type(job), 'cache_file', PropertyMock(return_value=cache_file))
    mocker.patch.object(job, '_is_executed', is_executed)
    open_mock = mocker.patch('upsies.jobs.base.open')
    job._write_cache()
    assert open_mock.call_args_list == []

def test_write_cache_writes_signal_emissions(tmp_path, mocker):
    class BarJob(FooJob):
        def execute(self):
            pass

    job = BarJob(home_directory=tmp_path, cache_directory=tmp_path)
    job.send('Foo')
    job.start()
    job.finish()
    mocker.patch.object(type(job.signal), 'emissions', PropertyMock(return_value='emissions mock'))
    job._write_cache()
    serialized_emissions = open(job.cache_file, 'rb').read()
    assert job._deserialize_from_cache(serialized_emissions) == 'emissions mock'

@pytest.mark.parametrize(
    argnames='exception, error',
    argvalues=(
        (OSError('No free space left'), 'No free space left'),
        (OSError(errno.EISDIR, 'Is directory'), 'Is directory'),
    ),
)
def test_write_cache_fails_to_write_cache_file(exception, error, job, mocker):
    open_mock = mocker.patch('upsies.jobs.base.open', side_effect=exception)
    mocker.patch.object(type(job.signal), 'emissions', PropertyMock(return_value='emissions mock'))
    mocker.patch.object(type(job), 'exit_code', PropertyMock(return_value=0))
    mocker.patch.object(type(job), 'cache_file', PropertyMock(return_value='path/to/cache'))
    mocker.patch.object(job, '_is_executed', True)
    with pytest.raises(RuntimeError, match=(rf'^Unable to write cache '
                                            rf'{re.escape(job.cache_file)}: {error}$')):
        job._write_cache()
    assert open_mock.call_args_list == [call(job.cache_file, 'wb')]

@pytest.mark.asyncio
async def test_cache_is_properly_written_when_repeating_command(tmp_path):
    class BarJob(FooJob):
        execute_counter = 0
        read_output_cache_counter = 0

        def execute(self):
            self.execute_counter += 1
            self.send('asdf')
            self.finish()

        def _read_cache(self):
            self.read_output_cache_counter += 1
            return super()._read_cache()

    for i in range(5):
        job = BarJob(home_directory=tmp_path, cache_directory=tmp_path)
        assert job.is_finished is False
        assert job.output == ()
        job.start()
        await job.wait()
        assert job.is_finished is True
        assert job.output == ('asdf',)
        assert job.execute_counter == (1 if i == 0 else 0)
        assert job.read_output_cache_counter == 1


@pytest.mark.parametrize(
    argnames='ignore_cache, cache_file, cache_file_exists',
    argvalues=(
        (True, 'path/to/cache_file', True),
        (1, 'path/to/cache_file', True),
        (False, '', True),
        (False, None, True),
        (1, 'path/to/cache_file', False),
    ),
)
def test_read_cache_does_nothing(ignore_cache, cache_file, cache_file_exists, job, mocker):
    mocker.patch.object(job, '_ignore_cache', ignore_cache)
    mocker.patch.object(type(job), 'cache_file', PropertyMock(return_value=cache_file))
    mocker.patch('os.path.exists', Mock(return_value=cache_file_exists))
    open_mock = mocker.patch('upsies.jobs.base.open')
    assert job._read_cache() is False
    assert open_mock.call_args_list == []

@pytest.mark.parametrize(
    argnames='exception, error',
    argvalues=(
        (OSError('No such file'), 'No such file'),
        (OSError(errno.EISDIR, 'Is directory'), 'Is directory'),
    ),
)
def test_read_cache_fails_to_read_cache_file(exception, error, job, mocker):
    mocker.patch('os.path.exists', return_value=True)
    open_mock = mocker.patch('upsies.jobs.base.open', side_effect=exception)
    with pytest.raises(RuntimeError, match=rf'^Unable to read cache {job.cache_file}: {error}$'):
        job._read_cache()
    assert open_mock.call_args_list == [call(job.cache_file, 'rb')]

def test_read_cache_replays_signal_emissions(job, mocker):
    open(job.cache_file, 'wb').write(job._serialize_for_cache('emissions mock'))
    replay_mock = mocker.patch.object(type(job.signal), 'replay')
    assert job._read_cache() is True
    assert replay_mock.call_args_list == [call('emissions mock')]

def test_read_cache_ignores_empty_signal_emission_cache(job, mocker):
    open(job.cache_file, 'wb').write(job._serialize_for_cache(''))
    replay_mock = mocker.patch.object(type(job.signal), 'replay')
    assert job._read_cache() is False
    assert replay_mock.call_args_list == []


def test_cache_file_when_cache_id_is_None(tmp_path, mocker):
    job = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    mocker.patch.object(type(job), 'cache_id', PropertyMock(return_value=None))
    assert job.cache_file is None

@pytest.mark.parametrize('cache_id_value', ('', {}, ()), ids=lambda v: repr(v))
def test_cache_file_when_cache_id_is_falsy(cache_id_value, tmp_path, mocker):
    job = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    mocker.patch.object(type(job), 'cache_id', PropertyMock(return_value=cache_id_value))
    assert job.cache_file == str(tmp_path / f'{job.name}.out')

def test_cache_file_when_cache_id_is_too_long(tmp_path, mocker):
    job = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    job._max_filename_len = 20
    long_cache_id = ''.join(str(n % 10) for n in range(job._max_filename_len))
    mocker.patch.object(type(job), 'cache_id', PropertyMock(return_value=long_cache_id))
    mocker.patch.object(type(job), '_cache_id_as_string', Mock(return_value=long_cache_id))
    assert job.cache_file == str(tmp_path / f'{job.name}.01234…56789.out')

def test_cache_file_when_cache_id_is_not_too_long(tmp_path, mocker):
    job = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    mocker.patch.object(type(job), 'cache_id', PropertyMock(return_value='something'))
    assert job.cache_file == str(tmp_path / f'{job.name}.something.out')

def test_cache_file_masks_illegal_characters(tmp_path, mocker):
    job = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    mocker.patch.object(type(job), 'cache_id', PropertyMock(return_value='hey you there'))
    mocker.patch('upsies.utils.fs.sanitize_filename', side_effect=lambda path: path.replace(' ', '#'))
    assert job.cache_file == str(tmp_path / f'{job.name}.hey#you#there.out')


def test_cache_id(tmp_path):
    job = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    assert job.cache_id == ''


def test_cache_id_as_string_when_cache_id_is_nonstring(mocker, tmp_path):
    job = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    assert job._cache_id_as_string(123) == '123'

def test_cache_id_as_string_when_cache_id_is_nonstring_iterable(mocker, tmp_path):
    job = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    assert job._cache_id_as_string((1, 2, 3)) == '1,2,3'

@pytest.mark.parametrize('value', (object(), print, lambda: None), ids=lambda v: str(v))
def test_cache_id_as_string_when_cache_id_is_mapping_with_nonstringable_key(value, mocker, tmp_path):
    job = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    with pytest.raises(RuntimeError, match=rf'^{re.escape(str(type(value)))} has no string representation$'):
        job._cache_id_as_string({1: 'foo', value: 'bar', 3: 'baz'})

@pytest.mark.parametrize('value', (object(), print, lambda: None), ids=lambda v: str(v))
def test_cache_id_as_string_when_cache_id_is_mapping_with_nonstringable_value(value, mocker, tmp_path):
    job = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    with pytest.raises(RuntimeError, match=rf'^{re.escape(str(type(value)))} has no string representation$'):
        job._cache_id_as_string({1: 'foo', 2: value, 3: 'baz'})

@pytest.mark.parametrize('value', (object(), print, lambda: None), ids=lambda v: str(v))
def test_cache_id_as_string_when_cache_id_is_sequence_with_nonstringable_item(value, mocker, tmp_path):
    job = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    with pytest.raises(RuntimeError, match=rf'^{re.escape(str(type(value)))} has no string representation$'):
        job._cache_id_as_string((1, value, 3))

def test_cache_id_as_string_normalizes_existing_paths(tmp_path):
    some_dir = tmp_path / 'foo' / 'bar'
    some_path = some_dir / 'baz.txt'
    some_dir.mkdir(parents=True)
    some_path.write_text('some thing')
    abs_path = some_path
    rel_path = some_path.relative_to(tmp_path)
    orig_cwd = os.getcwd()
    os.chdir(tmp_path)
    try:
        job1 = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
        job2 = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
        assert job1._cache_id_as_string(abs_path) == job2._cache_id_as_string(rel_path)
    finally:
        os.chdir(orig_cwd)

def test_cache_id_as_string_normalizes_multibyte_characters(tmp_path, mocker):
    job = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    mocker.patch.object(type(job), 'cache_id', PropertyMock(return_value='kožušček'))
    assert job.cache_file == str(tmp_path / f'{job.name}.kozuscek.out')



import os
import re
import time
from unittest.mock import AsyncMock, Mock, PropertyMock, call, patch

import pytest

from upsies.utils import webdbs
from upsies.utils.release import Episodes, ReleaseInfo, ReleaseName
from upsies.utils.types import ReleaseType


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_name_argument(ReleaseInfo_mock, mocker):
    rn = ReleaseName('path/to/something', name='the real something')
    assert ReleaseInfo_mock.call_args_list == [call('the real something')]
    assert rn._info is ReleaseInfo_mock.return_value


def test_release_info_attribute():
    path = 'path/to/Foo Season 1 1080p BluRay DTS-ASDF/Foo S01E02 Uncut x264.mkv'
    rn = ReleaseName(path)
    assert isinstance(rn.release_info, ReleaseInfo)
    assert rn.release_info == ReleaseInfo(path)


def test_set_release_info(mocker):
    mocker.patch('upsies.utils.video.audio_format', return_value='DTS')
    rn = ReleaseName('path/to/Foo Season 1 1080p BluRay DTS-ASDF/Foo S01E02 Uncut x264.mkv')
    assert str(rn) == 'Foo S01E02 Uncut 1080p BluRay DTS x264-ASDF'
    rn.set_release_info('path/to/Bar Season 2 720p Limited WEBDL FLAC-AsdF/Bar S02E03 x265.mkv')
    assert str(rn) == 'Bar S02E03 Limited 720p WEB-DL DTS x265-AsdF'


@patch('upsies.utils.release.ReleaseInfo', Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='path, name, translate, exp_repr',
    argvalues=(
        ('path/to/something', None, None,
         "ReleaseName('path/to/something')"),
        ('path/to/something', 'The Name', None,
         "ReleaseName('path/to/something', name='The Name')"),
        ('path/to/something', 'The Name', {'foo': 'bar'},
         "ReleaseName('path/to/something', name='The Name', translate={'foo': 'bar'})"),
        ('path/to/something', None, {'foo': 'bar'},
         "ReleaseName('path/to/something', translate={'foo': 'bar'})"),
    ),
)
def test_repr(path, name, translate, exp_repr):
    rn = ReleaseName(path=path, name=name, translate=translate)
    assert repr(rn) == exp_repr


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_str(ReleaseInfo_mock):
    rn = ReleaseName('path/to/something')
    with patch.object(rn, 'format') as format_mock:
        format_mock.return_value = 'Pretty Release Name'
        assert str(rn) == 'Pretty Release Name'
        format_mock.call_args_list == [call()]


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='attr',
    argvalues=tuple(ReleaseName('')),
)
def test_getitem(ReleaseInfo_mock, attr):
    ReleaseInfo_mock.return_value = {attr: 'mock value', 'type': ReleaseType.movie}
    rn = ReleaseName('path/to/something')
    if not attr.startswith('_'):
        rn[attr]
    with pytest.raises(KeyError, match=r"^'format'$"):
        rn['format']
    with pytest.raises(TypeError, match=r"^Not a string: 123$"):
        rn[123]


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_usage_as_dictionary_in_str_format(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {
        'type': ReleaseType.movie,
        'title': 'The Foo',
        'year': '1998',
        'source': 'BluRay',
        'resolution': '1080p',
        'audio_codec': 'AC3',
        'audio_channels': '5.1',
        'video_codec': 'x264',
        'group': 'ASDF',
    }
    rn = ReleaseName('path/to/something')
    fmt = '{title} ({year}) {audio_format} {video_format} {source}-{group}'
    assert fmt.format(**rn) == 'The Foo (1998) AC3 x264 BluRay-ASDF'


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_path(ReleaseInfo_mock):
    rn = ReleaseName('path/to/something')
    assert rn.path == 'path/to/something'
    rn = ReleaseName(123)
    assert rn.path == '123'


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_separator(ReleaseInfo_mock):
    rn = ReleaseName('path/to/something')
    assert rn.separator == ' '
    rn.separator = 'asdf'
    assert rn.separator == 'asdf'
    rn.separator = 123
    assert rn.separator == '123'
    rn.separator = None
    assert rn.separator == ' '


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_type_getter(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'type': ReleaseType.movie}
    assert ReleaseName('path/to/something').type is ReleaseType.movie
    ReleaseInfo_mock.return_value = {'type': ReleaseType.season}
    assert ReleaseName('path/to/something').type is ReleaseType.season
    ReleaseInfo_mock.return_value = {'type': ReleaseType.episode}
    assert ReleaseName('path/to/something').type is ReleaseType.episode
    ReleaseInfo_mock.return_value = {}
    assert ReleaseName('path/to/something').type is ReleaseType.unknown

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='value, exp_value',
    argvalues=(
        ('movie', ReleaseType.movie),
        ('season', ReleaseType.season),
        ('episode', ReleaseType.episode),
        ('', ReleaseType.unknown),
        (ReleaseType.movie, ReleaseType.movie),
        (ReleaseType.season, ReleaseType.season),
        (ReleaseType.episode, ReleaseType.episode),
        (ReleaseType.unknown, ReleaseType.unknown),
    ),
)
def test_type_setter_with_valid_value(ReleaseInfo_mock, value, exp_value):
    rn = ReleaseName('path/to/something')
    rn.type = value
    assert rn.type is exp_value

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_type_setter_with_invalid_value(ReleaseInfo_mock):
    rn = ReleaseName('path/to/something')
    with pytest.raises(ValueError, match=r"^'asdf' is not a valid ReleaseType$"):
        rn.type = 'asdf'


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_title_getter(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'title': 'The Foo'}
    assert ReleaseName('path/to/something').title == 'The Foo'
    ReleaseInfo_mock.return_value = {'title': 'The Bar'}
    assert ReleaseName('path/to/something').title == 'The Bar'

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_title_setter(ReleaseInfo_mock):
    rn = ReleaseName('path/to/something')
    assert rn.title != 'The Baz'
    rn.title = 'The Baz'
    assert rn.title == 'The Baz'

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_title_is_translated(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'title': 'The Foo', 'aka': 'Bar', 'year': '2015'}
    translation = {
        'title': {
            re.compile(r'o+'): r'00',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    rn.year_required = True
    assert rn.title == 'The F00'
    assert rn.title_aka == 'Bar'
    assert rn.title_with_aka == 'The F00 AKA Bar'
    assert rn.title_with_aka_and_year == 'The F00 AKA Bar 2015'


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_title_aka_getter(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'title': 'The Foo', 'aka': ''}
    assert ReleaseName('path/to/something').title_aka == ''
    ReleaseInfo_mock.return_value = {'title': 'The Foo', 'aka': 'The Bar'}
    assert ReleaseName('path/to/something').title_aka == 'The Bar'
    ReleaseInfo_mock.return_value = {'title': 'The Foo', 'aka': 'The Foo'}
    assert ReleaseName('path/to/something').title_aka == ''

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_title_aka_setter(ReleaseInfo_mock):
    rn = ReleaseName('path/to/something')
    assert rn.title_aka != 'The Baz'
    rn.title_aka = 'The Baz'
    assert rn.title_aka == 'The Baz'
    rn.title = 'The Baz'
    assert rn.title_aka == ''

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_title_aka_is_translated(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'title': 'The Foo', 'aka': 'Bar', 'year': '2015'}
    translation = {
        'title_aka': {
            re.compile(r'(\w+)r'): r'R-\1',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    rn.year_required = True
    assert rn.title == 'The Foo'
    assert rn.title_aka == 'R-Ba'
    assert rn.title_with_aka == 'The Foo AKA R-Ba'
    assert rn.title_with_aka_and_year == 'The Foo AKA R-Ba 2015'


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_title_with_aka(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'title': 'The Foo', 'aka': ''}
    assert ReleaseName('path/to/something').title_with_aka == 'The Foo'
    ReleaseInfo_mock.return_value = {'title': 'The Foo', 'aka': 'The Bar'}
    assert ReleaseName('path/to/something').title_with_aka == 'The Foo AKA The Bar'

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_title_with_aka_is_translated(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'title': 'The Foo', 'aka': 'Bar', 'year': '2015'}
    translation = {
        'title_with_aka': {
            re.compile(r' +AKA +'): r': ',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    rn.year_required = True
    assert rn.title == 'The Foo'
    assert rn.title_aka == 'Bar'
    assert rn.title_with_aka == 'The Foo: Bar'
    assert rn.title_with_aka_and_year == 'The Foo: Bar 2015'


@pytest.mark.parametrize(
    argnames='date, year, country, year_required, country_required, exp_title_with_aka_and_year',
    argvalues=(
        ('', '2003', 'UK', False, False, 'The Foo'),
        ('', '2003', 'UK', False, True, 'The Foo UK'),
        ('', '2003', 'UK', False, False, 'The Foo'),
        ('', '2003', 'UK', True, False, 'The Foo 2003'),
        ('', '2003', 'UK', True, True, 'The Foo 2003 UK'),

        ('1999-10-03', '2003', 'UK', False, False, 'The Foo 1999-10-03'),
        ('1999-10-03', '2003', 'UK', False, True, 'The Foo 1999-10-03 UK'),
        ('1999-10-03', '2003', 'UK', False, False, 'The Foo 1999-10-03'),
        ('1999-10-03', '2003', 'UK', True, False, 'The Foo 1999-10-03'),
        ('1999-10-03', '2003', 'UK', True, True, 'The Foo 1999-10-03 UK'),
    ),
    ids=lambda v: str(v),
)
def test_title_with_aka_and_year(date, year, country, year_required, country_required, exp_title_with_aka_and_year, mocker):
    rn = ReleaseName('path/to/something')

    mocker.patch.object(type(rn), 'title_with_aka', PropertyMock(return_value='The Foo'))
    mocker.patch.object(type(rn), 'date', PropertyMock(return_value=date))
    mocker.patch.object(type(rn), 'year', PropertyMock(return_value=year))
    mocker.patch.object(type(rn), 'year_required', PropertyMock(return_value=year_required))
    mocker.patch.object(type(rn), 'country', PropertyMock(return_value=country))
    mocker.patch.object(type(rn), 'country_required', PropertyMock(return_value=country_required))

    rn.year_required = year_required
    rn.country_required = country_required
    assert rn.title_with_aka_and_year == exp_title_with_aka_and_year

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_title_with_aka_and_year_is_translated(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'title': 'The Foo', 'aka': 'Bar', 'year': '2015'}
    translation = {
        'title_with_aka_and_year': {
            re.compile(r'(.*) (\d+)'): r'"\1" \2',
            re.compile(r'(\d+)'): r'(\1)',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    rn.year_required = True
    assert rn.title == 'The Foo'
    assert rn.title_aka == 'Bar'
    assert rn.title_with_aka == 'The Foo AKA Bar'
    assert rn.title_with_aka_and_year == '"The Foo AKA Bar" (2015)'


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_year_getter_with_movie(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'year': '2000', 'type': ReleaseType.movie}
    assert ReleaseName('path/to/something').year == '2000'
    ReleaseInfo_mock.return_value = {'year': '', 'type': ReleaseType.movie}
    assert ReleaseName('path/to/something').year == 'UNKNOWN_YEAR'
    ReleaseInfo_mock.return_value = {'type': ReleaseType.movie}
    assert ReleaseName('path/to/something').year == 'UNKNOWN_YEAR'

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize('type', (ReleaseType.season, ReleaseType.episode))
def test_year_getter_with_series(ReleaseInfo_mock, type):
    ReleaseInfo_mock.return_value = {'year': '2000', 'type': type}
    assert ReleaseName('path/to/something').year == '2000'
    ReleaseInfo_mock.return_value = {'year': '', 'type': type}
    assert ReleaseName('path/to/something').year == ''
    ReleaseInfo_mock.return_value = {'type': type}
    assert ReleaseName('path/to/something').year == ''

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='year, exp_year',
    argvalues=(
        ('1880', '1880'),
        ('1990', '1990'),
        ('2015', '2015'),
        (time.strftime('%Y'), time.strftime('%Y')),
        (int(time.strftime('%Y')) + 1, str(int(time.strftime('%Y')) + 1)),
        (int(time.strftime('%Y')) + 2, str(int(time.strftime('%Y')) + 2)),
        ('', ''),
        (None, ''),
    ),
)
def test_year_setter_with_valid_year(ReleaseInfo_mock, year, exp_year):
    rn = ReleaseName('path/to/something')
    assert rn.year == ''
    rn.year = year
    assert rn.year == exp_year

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='year, exp_exception, exp_message',
    argvalues=(
        ([], TypeError, 'Not a number: []'),
        ('foo', ValueError, 'Invalid year: foo'),
        ('1879', ValueError, 'Invalid year: 1879'),
        (int(time.strftime('%Y')) + 3, ValueError, f'Invalid year: {int(time.strftime("%Y")) + 3}'),
    ),
)
def test_year_setter_with_invalid_year(ReleaseInfo_mock, year, exp_exception, exp_message):
    rn = ReleaseName('path/to/something')
    assert rn.year == ''
    with pytest.raises(exp_exception, match=rf'^{re.escape(exp_message)}$'):
        rn.year = year

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_year_is_translated(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'year': '2012'}
    translation = {
        'year': {
            re.compile(r'(\d+)'): r'\1 CE',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    assert rn.year == '2012 CE'
    rn.year = '1995'
    assert rn.year == '1995 CE'


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='year_required, year, exp_year',
    argvalues=(
        (False, '', ''),
        (True, '', 'UNKNOWN_YEAR'),
        (False, '1234', '1234'),
        (True, '1234', '1234'),
    ),
)
def test_year_required(ReleaseInfo_mock, year_required, year, exp_year):
    ReleaseInfo_mock.return_value = {'year': year}
    rn = ReleaseName('path/to/something')
    rn.year_required = year_required
    assert rn.year_required is year_required
    assert rn.year == exp_year


@pytest.mark.parametrize('country_required', (False, True))
@pytest.mark.parametrize('country', (None, '', 'UK'))
def test_country_getter(country, country_required, mocker):
    mocker.patch('upsies.utils.release.ReleaseInfo', Mock(return_value={
        'country': country,
    }))
    rn = ReleaseName('path/to/something')
    rn.country_required = country_required
    if country_required:
        if country:
            assert rn.country == country
        else:
            assert rn.country == 'UNKNOWN_COUNTRY'
    else:
        if country:
            assert rn.country == country
        else:
            assert rn.country == ''

@pytest.mark.parametrize(
    argnames='country, exp_country',
    argvalues=(
        ('UK', 'UK'),
        ([1, 2, 3], '[1, 2, 3]'),
        ('', ''),
        (None, ''),
    ),
)
def test_country_setter(country, exp_country, mocker):
    mocker.patch('upsies.utils.release.ReleaseInfo', Mock(return_value={}))
    rn = ReleaseName('path/to/something')
    assert rn.country == ''
    rn.country = country
    assert rn.country == exp_country

def test_country_is_translated(mocker):
    mocker.patch('upsies.utils.release.ReleaseInfo', Mock(return_value={
        'country': 'UK',
    }))
    translation = {
        'country': {
            re.compile(r'^(.+)$'): r'[\1]',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    assert rn.country == '[UK]'
    rn.country = 'US'
    assert rn.country == '[US]'


def test_country_required(mocker):
    mocker.patch('upsies.utils.release.ReleaseInfo', Mock(return_value={}))
    rn = ReleaseName('path/to/something')
    assert not hasattr(rn, '_country_required')
    assert rn.country_required is False
    rn.country_required = 1
    assert rn.country_required is True
    rn.country_required = 0
    assert rn.country_required is False


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='type, episodes, exp_string',
    argvalues=(
        (ReleaseType.movie, Episodes({}), ''),
        (ReleaseType.movie, Episodes({3: ()}), ''),
        (ReleaseType.movie, Episodes({3: (1, 2), 4: ()}), ''),
        (ReleaseType.movie, Episodes({3: (1, 2), 4: (3, 4)}), ''),

        (ReleaseType.season, Episodes({}), 'UNKNOWN_SEASON'),
        (ReleaseType.season, Episodes({3: ()}), 'S03'),
        (ReleaseType.season, Episodes({3: (), 4: ()}), 'S03S04'),
        (ReleaseType.season, Episodes({3: (1, 2), 4: ()}), 'S03E01E02S04'),
        (ReleaseType.season, Episodes({3: (1, 2), 4: (3, 4)}), 'S03E01E02S04E03E04'),

        (ReleaseType.episode, Episodes({}), 'UNKNOWN_EPISODE'),
        (ReleaseType.episode, Episodes({3: ()}), 'UNKNOWN_EPISODE'),
        (ReleaseType.episode, Episodes({3: (), 4: ()}), 'UNKNOWN_EPISODE'),
        (ReleaseType.episode, Episodes({3: (1, 2), 4: ()}), 'S03E01E02S04'),
        (ReleaseType.episode, Episodes({3: (1, 2), 4: (3, 4)}), 'S03E01E02S04E03E04'),

        (ReleaseType.unknown, Episodes({}), ''),
        (ReleaseType.unknown, Episodes({3: ()}), 'S03'),
        (ReleaseType.unknown, Episodes({3: (), 4: ()}), 'S03S04'),
        (ReleaseType.unknown, Episodes({3: (1, 2), 4: ()}), 'S03E01E02S04'),
        (ReleaseType.unknown, Episodes({3: (1, 2), 4: (3, 4)}), 'S03E01E02S04E03E04'),
    ),
)
def test_episodes_getter(ReleaseInfo_mock, type, episodes, exp_string):
    ReleaseInfo_mock.return_value = {'episodes': episodes, 'type': type}
    assert ReleaseName('path/to/something').episodes == exp_string

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='type, value, exp_episodes',
    argvalues=(
        (ReleaseType.movie, 'foo.S03S04.bar', ''),
        (ReleaseType.movie, '3', ''),
        (ReleaseType.movie, 3, ''),
        (ReleaseType.movie, (3, '4'), ''),
        (ReleaseType.movie, {3: (1, 2), '4': ()}, ''),
        (ReleaseType.movie, '', ''),
        (ReleaseType.season, 'foo.S03S04.bar', 'S03S04'),
        (ReleaseType.season, 'foo.S03E10.bar', 'S03E10'),
        (ReleaseType.season, '3', 'S03'),
        (ReleaseType.season, 3, 'S03'),
        (ReleaseType.season, (3, '4'), 'S03S04'),
        (ReleaseType.season, {3: (1, 2), '4': ()}, 'S03E01E02S04'),
        (ReleaseType.season, '', 'UNKNOWN_SEASON'),
        (ReleaseType.episode, 'foo.S03S04.bar', 'UNKNOWN_EPISODE'),
        (ReleaseType.episode, 'foo.S03E10.bar', 'S03E10'),
        (ReleaseType.episode, 'foo.S03E10E11.bar', 'S03E10E11'),
        (ReleaseType.episode, '3', 'UNKNOWN_EPISODE'),
        (ReleaseType.episode, 3, 'UNKNOWN_EPISODE'),
        (ReleaseType.episode, (3, '4'), 'UNKNOWN_EPISODE'),
        (ReleaseType.episode, {3: (1, 2), '4': ()}, 'S03E01E02S04'),
        (ReleaseType.episode, '', 'UNKNOWN_EPISODE'),
        (ReleaseType.unknown, 'foo.S03S04.bar', 'S03S04'),
        (ReleaseType.unknown, 'foo.S03E10.bar', 'S03E10'),
        (ReleaseType.unknown, 'foo.S03E10E11.bar', 'S03E10E11'),
        (ReleaseType.unknown, '3', 'S03'),
        (ReleaseType.unknown, 3, 'S03'),
        (ReleaseType.unknown, (3, '4'), 'S03S04'),
        (ReleaseType.unknown, {3: (1, 2), '4': ()}, 'S03E01E02S04'),
        (ReleaseType.unknown, '', ''),
    ),
)
def test_episodes_setter_with_valid_value(ReleaseInfo_mock, type, value, exp_episodes):
    rn = ReleaseName('path/to/something')
    rn.type = type
    rn.episodes = value
    assert rn.episodes == exp_episodes

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_episodes_is_translated(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'episodes': Episodes({3: ()})}
    translation = {
        'episodes': {
            re.compile(r'S0*'): r'Season ',
            re.compile(r'E0*'): r', Episode ',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    assert rn.episodes == 'Season 3'
    rn.episodes = 'S03E12'
    assert rn.episodes == 'Season 3, Episode 12'


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_episodes_dict(ReleaseInfo_mock):
    rn = ReleaseName('path/to/something')
    assert rn.episodes_dict == {}
    rn.episodes = 'S03E12S04E13S05S06E12E20'
    assert rn.episodes_dict == {
        '3': ['12'],
        '4': ['13'],
        '5': [],
        '6': ['12', '20'],
    }


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_only_season(ReleaseInfo_mock):
    rn = ReleaseName('path/to/something')
    assert rn.only_season is None
    rn.episodes = 'S03E12S04E13S05S06E12E20'
    assert rn.only_season is None
    rn.episodes = 'S02'
    assert rn.only_season == '2'
    rn.episodes = 'S02S010'
    assert rn.only_season is None


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='type, episode_title, exp_episode_title',
    argvalues=(
        (ReleaseType.movie, 'Something', ''),
        (ReleaseType.movie, '', ''),
        (ReleaseType.season, 'Something', ''),
        (ReleaseType.season, '', ''),
        (ReleaseType.episode, 'Something', 'Something'),
        (ReleaseType.episode, '', ''),
        (ReleaseType.unknown, 'Something', ''),
        (ReleaseType.unknown, '', ''),
    ),
)
def test_episode_title_getter(ReleaseInfo_mock, type, episode_title, exp_episode_title):
    ReleaseInfo_mock.return_value = {'episode_title': episode_title, 'type': type}
    rn = ReleaseName('path/to/something')
    assert rn.episode_title == exp_episode_title
    rn.type = ReleaseType.episode
    assert rn.episode_title == episode_title

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize('episode_title, exp_episode_title', (('Foo', 'Foo'), (123, '123')))
def test_episode_title_setter(ReleaseInfo_mock, episode_title, exp_episode_title):
    rn = ReleaseName('path/to/something')
    rn.type = ReleaseType.episode
    rn.episode_title = episode_title
    assert rn.episode_title == exp_episode_title

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_episode_title_is_translated(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'episode_title': 'This Episode', 'type': ReleaseType.episode}
    translation = {
        'episode_title': {
            re.compile(r'(?i:this)'): r'That',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    assert rn.episode_title == 'That Episode'
    rn.episode_title = 'foo'
    assert rn.episode_title == 'foo'


@pytest.mark.parametrize(
    argnames='release_type, date, exp_date',
    argvalues=(
        (ReleaseType.unknown, '2000-10-20', ''),
        (ReleaseType.movie, '2000-10-20', ''),
        (ReleaseType.season, '2000-10-20', ''),
        (ReleaseType.episode, '2000-10-20', '2000-10-20'),
        (ReleaseType.episode, None, ''),
    ),
)
@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_date_getter(ReleaseInfo_mock, release_type, date, exp_date):
    ReleaseInfo_mock.return_value = {'type': release_type, 'date': date}
    assert ReleaseName('path/to/something').date == exp_date

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_date_setter(ReleaseInfo_mock):
    rn = ReleaseName('path/to/something')
    rn.type = ReleaseType.episode
    assert rn.date == ''
    rn.date = '1234-56-78'
    assert rn.date == '1234-56-78'

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_date_is_translated(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'type': ReleaseType.episode}
    translation = {
        'date': {
            re.compile(r'(\d{4})-(\d{2})-(\d{2})'): r'\3-\2-\1',
            re.compile(r'-'): r'.',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    rn.date = '1234-56-78'
    assert rn.date == '78.56.1234'


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_service_getter(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'service': 'FOO'}
    assert ReleaseName('path/to/something').service == 'FOO'

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_service_setter(ReleaseInfo_mock):
    rn = ReleaseName('path/to/something')
    assert rn.service == ''
    rn.service = 'FOO'
    assert rn.service == 'FOO'
    rn.service = 256
    assert rn.service == '256'

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_service_is_translated(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'service': 'Asdf'}
    translation = {
        'service': {
            re.compile(r'Asdf'): r'ASDF',
            re.compile(r'(.*)TV'): r'Tee\1Vee',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    assert rn.service == 'ASDF'
    rn.service = 'AsdfTV'
    assert rn.service == 'TeeASDFVee'


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_edition_getter_returns_same_list_with_no_edition(ReleaseInfo_mock):
    rn = ReleaseName('path/to/something')
    assert rn.edition is rn.edition
    assert isinstance(rn.edition, list)

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_edition_getter_returns_same_list_with_given_edition(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'edition': ["Director's Cut", 'Unrated']}
    rn = ReleaseName('path/to/something')
    assert rn.edition == ["Director's Cut", 'Unrated']
    assert rn.edition is rn.edition
    assert isinstance(rn.edition, list)

@patch('upsies.utils.release.ReleaseInfo', Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='edition_list, has_dual_audio, exp_edition',
    argvalues=(
        ([], True, ['Dual Audio']),
        ([], False, []),
        (['Dual Audio'], True, ['Dual Audio']),
        (['Dual Audio'], False, []),
        (['Dual Audio', 'Dual Audio'], False, []),
    ),
)
def test_edition_getter_autodetects_dual_audio(edition_list, has_dual_audio, exp_edition, mocker):
    rn = ReleaseName('path/to/something')
    rn._info['edition'] = edition_list
    mocker.patch.object(type(rn), 'has_dual_audio', PropertyMock(return_value=has_dual_audio))
    # Ensure "Dual Audio" is only appended once
    for _ in range(3):
        assert rn.edition == exp_edition

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='edition, exp_edition',
    argvalues=(
        (('Special Edition',), ['Special Edition']),
        ((1, 2, 3), ['1', '2', '3']),
    ),
)
def test_edition_setter(ReleaseInfo_mock, edition, exp_edition):
    rn = ReleaseName('path/to/something')
    rn.edition = edition
    assert rn.edition == exp_edition
    assert rn.edition is rn.edition

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_edition_is_translated(ReleaseInfo_mock, mocker):
    ReleaseInfo_mock.return_value = {'edition': ['foo', 'bar', 'baz']}
    rn = ReleaseName('path/to/something')
    translation = {
        'edition': {
            re.compile(r'o+'): r'00',
            re.compile(r'b(\w+)'): r'\1F',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    assert rn.edition == ['f00', 'arF', 'azF']
    rn.edition = ('noo', 'nar', 'naz')
    assert rn.edition == ['n00', 'nar', 'naz']


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='source, exp_source',
    argvalues=(
        ('WEB-DL', 'WEB-DL'),
        ('', 'UNKNOWN_SOURCE'),
    ),
)
def test_source_getter(ReleaseInfo_mock, source, exp_source):
    ReleaseInfo_mock.return_value = {'source': source}
    assert ReleaseName('path/to/something').source == exp_source

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='video_ts_size, exp_source',
    argvalues=(
        (4.7e9, 'DVD5'),
        (4.7e9 + 1, 'DVD9'),
    ),
)
def test_source_getter_with_video_ts(ReleaseInfo_mock, video_ts_size, exp_source, mocker):
    ReleaseInfo_mock.return_value = {'source': 'Ignored'}
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch('upsies.utils.fs.find_name', return_value='path/to/something/VIDEO_TS'),
        'find_name',
    )
    mocks.attach_mock(
        mocker.patch('upsies.utils.fs.path_size', return_value=video_ts_size),
        'path_size',
    )
    assert ReleaseName('path/to/something').source == exp_source
    assert mocks.mock_calls == [
        call.find_name('VIDEO_TS', 'path/to/something', validator=os.path.isdir),
        call.path_size('path/to/something'),
    ]

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='bdmv_size, exp_source',
    argvalues=(
        (25e9 + 0, 'BD25'),
        (25e9 + 1, 'BD50'),
        (50e9 + 0, 'BD50'),
        (50e9 + 1, 'BD66'),
        (66e9 + 0, 'BD66'),
        (66e9 + 1, 'BD100'),
    ),
)
def test_source_getter_with_bdmv(ReleaseInfo_mock, bdmv_size, exp_source, mocker):
    ReleaseInfo_mock.return_value = {'source': 'Ignored'}
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch('upsies.utils.fs.find_name', side_effect=[None, 'path/to/something/BDMV']),
        'find_name',
    )
    mocks.attach_mock(
        mocker.patch('upsies.utils.fs.path_size', return_value=bdmv_size),
        'path_size',
    )
    assert ReleaseName('path/to/something').source == exp_source
    assert mocks.mock_calls == [
        call.find_name('VIDEO_TS', 'path/to/something', validator=os.path.isdir),
        call.find_name('BDMV', 'path/to/something', validator=os.path.isdir),
        call.path_size('path/to/something'),
    ]

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@pytest.mark.parametrize(
    argnames='source, exp_source',
    argvalues=(
        ('BluRay', 'BluRay'),
        ('', 'UNKNOWN_SOURCE'),
        (123, '123'),
    ),
)
def test_source_setter(ReleaseInfo_mock, source, exp_source):
    rn = ReleaseName('path/to/something')
    rn.source = source
    assert rn.source == exp_source

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_source_is_translated(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'source': 'Blu-Ray'}
    translation = {
        'source': {
            re.compile(r'-'): r'',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    assert rn.source == 'BluRay'
    rn.source = 'asdf'
    assert rn.source == 'asdf'
    rn.source = 'WEB-DL'
    assert rn.source == 'WEBDL'


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@patch('upsies.utils.video.resolution')
def test_resolution_getter_prefers_mediainfo(resolution_mock, ReleaseInfo_mock):
    resolution_mock.return_value = '720p'
    ReleaseInfo_mock.return_value = {'resolution': '1080p'}
    assert ReleaseName('path/to/something').resolution == '720p'
    assert resolution_mock.call_args_list == [call('path/to/something', default=None)]

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@patch('upsies.utils.video.resolution')
def test_resolution_getter_defaults_to_guess(resolution_mock, ReleaseInfo_mock):
    resolution_mock.return_value = None
    ReleaseInfo_mock.return_value = {'resolution': '1080p'}
    assert ReleaseName('path/to/something').resolution == '1080p'
    assert resolution_mock.call_args_list == [call('path/to/something', default=None)]

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@patch('upsies.utils.video.resolution')
def test_resolution_getter_defaults_to_placeholder(resolution_mock, ReleaseInfo_mock):
    resolution_mock.return_value = None
    ReleaseInfo_mock.return_value = {}
    assert ReleaseName('path/to/something').resolution == 'UNKNOWN_RESOLUTION'
    assert resolution_mock.call_args_list == [call('path/to/something', default=None)]

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_resolution_setter(ReleaseInfo_mock):
    rn = ReleaseName('path/to/something')
    assert rn.resolution == 'UNKNOWN_RESOLUTION'
    rn.resolution = '1080p'
    assert rn.resolution == '1080p'
    rn.resolution = ''
    assert rn.resolution == 'UNKNOWN_RESOLUTION'
    rn.resolution = 1080
    assert rn.resolution == '1080'

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
@patch('upsies.utils.video.resolution')
def test_resolution_is_translated(resolution_mock, ReleaseInfo_mock):
    resolution_mock.return_value = None
    ReleaseInfo_mock.return_value = {'resolution': '1080i'}
    translation = {
        'resolution': {
            re.compile(r'(\d+)i'): r'\1I',
            re.compile(r'x'): r':',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    assert rn.resolution == '1080I'
    rn.resolution = 'asdf'
    assert rn.resolution == 'asdf'
    rn.resolution = '1080x1920'
    assert rn.resolution == '1080:1920'


@patch('upsies.utils.release.ReleaseInfo')
@patch('upsies.utils.video.audio_format')
def test_audio_format_getter_prefers_mediainfo(audio_format_mock, ReleaseInfo_mock):
    audio_format_mock.return_value = 'AC3'
    ReleaseInfo_mock.return_value = {'audio_codec': 'DD+'}
    assert ReleaseName('path/to/something').audio_format == 'AC3'
    assert audio_format_mock.call_args_list == [call('path/to/something', default=None)]

@patch('upsies.utils.release.ReleaseInfo')
@patch('upsies.utils.video.audio_format')
def test_audio_format_getter_defaults_to_guess(audio_format_mock, ReleaseInfo_mock):
    audio_format_mock.return_value = None
    ReleaseInfo_mock.return_value = {'audio_codec': 'DD+'}
    assert ReleaseName('path/to/something').audio_format == 'DD+'
    assert audio_format_mock.call_args_list == [call('path/to/something', default=None)]

@patch('upsies.utils.release.ReleaseInfo')
@patch('upsies.utils.video.audio_format')
def test_audio_format_getter_defaults_to_empty_string(audio_format_mock, ReleaseInfo_mock):
    audio_format_mock.return_value = None
    ReleaseInfo_mock.return_value = {'audio_codec': ''}
    assert ReleaseName('path/to/something').audio_format == ''
    assert audio_format_mock.call_args_list == [call('path/to/something', default=None)]

@patch('upsies.utils.release.ReleaseInfo')
def test_audio_format_setter(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {}
    rn = ReleaseName('path/to/something')
    assert rn.audio_format == ''
    rn.audio_format = 'AC3'
    assert rn.audio_format == 'AC3'
    rn.audio_format = ''
    assert rn.audio_format == ''
    rn.audio_format = 'DD+'
    assert rn.audio_format == 'DD+'

@patch('upsies.utils.release.ReleaseInfo')
@patch('upsies.utils.video.audio_format')
def test_audio_format_is_translated(audio_format_mock, ReleaseInfo_mock):
    audio_format_mock.return_value = None
    ReleaseInfo_mock.return_value = {'audio_codec': 'fooo'}
    translation = {
        'audio_format': {
            re.compile(r'\w(o+)'): r'b\1',
            re.compile(r'oo'): r'00',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    assert rn.audio_format == 'b00o'
    rn.audio_format = 'asdf'
    assert rn.audio_format == 'asdf'
    rn.audio_format = 'noooo'
    assert rn.audio_format == 'b0000'


@patch('upsies.utils.release.ReleaseInfo')
@patch('upsies.utils.video.audio_channels')
def test_audio_channels_getter_prefers_mediainfo(audio_channels_mock, ReleaseInfo_mock):
    audio_channels_mock.return_value = '5.1'
    ReleaseInfo_mock.return_value = {'audio_channels': '7.1'}
    assert ReleaseName('path/to/something').audio_channels == '5.1'
    assert audio_channels_mock.call_args_list == [call('path/to/something', default=None)]

@patch('upsies.utils.release.ReleaseInfo')
@patch('upsies.utils.video.audio_channels')
def test_audio_channels_getter_defaults_to_guess(audio_channels_mock, ReleaseInfo_mock):
    audio_channels_mock.return_value = None
    ReleaseInfo_mock.return_value = {'audio_channels': '7.1'}
    assert ReleaseName('path/to/something').audio_channels == '7.1'
    assert audio_channels_mock.call_args_list == [call('path/to/something', default=None)]

@patch('upsies.utils.release.ReleaseInfo')
@patch('upsies.utils.video.audio_channels')
def test_audio_channels_getter_defaults_to_empty_string(audio_channels_mock, ReleaseInfo_mock):
    audio_channels_mock.return_value = None
    ReleaseInfo_mock.return_value = {}
    assert ReleaseName('path/to/something').audio_channels == ''
    assert audio_channels_mock.call_args_list == [call('path/to/something', default=None)]

@patch('upsies.utils.release.ReleaseInfo')
def test_audio_channels_setter(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {}
    rn = ReleaseName('path/to/something')
    assert rn.audio_channels == ''
    rn.audio_channels = '2.0'
    assert rn.audio_channels == '2.0'
    rn.audio_channels = ''
    assert rn.audio_channels == ''
    rn.audio_channels = 2.0
    assert rn.audio_channels == '2.0'

@patch('upsies.utils.release.ReleaseInfo')
@patch('upsies.utils.video.audio_channels')
def test_audio_channels_is_translated(audio_channels_mock, ReleaseInfo_mock):
    audio_channels_mock.return_value = None
    ReleaseInfo_mock.return_value = {'audio_channels': '5.1'}
    translation = {'audio_channels': {re.compile(r'\.'): r':'}}
    rn = ReleaseName('path/to/something', translate=translation)
    assert rn.audio_channels == '5:1'
    rn.audio_channels = 'asdf'
    assert rn.audio_channels == 'asdf'
    rn.audio_channels = '2.0'
    assert rn.audio_channels == '2:0'


@pytest.mark.parametrize(
    argnames='_hdr_format, path_exists, hdr_from_mediainfo, hdr_from_release_info, exp_value',
    argvalues=(
        ('custom HDR format', 'path_does_not_exist', 'HDR [mediainfo]', 'HDR [release info]', 'custom HDR format'),
        ('custom HDR format', 'path_does_exist', 'HDR [mediainfo]', 'HDR [release info]', 'custom HDR format'),
        ('', 'path_does_not_exist', 'HDR [mediainfo]', 'HDR [release info]', ''),
        ('', 'path_does_exist', 'HDR [mediainfo]', 'HDR [release info]', ''),
        (None, 'path_does_not_exist', 'HDR [mediainfo]', 'HDR [release info]', 'HDR [release info]'),
        (None, 'path_does_exist', 'HDR [mediainfo]', 'HDR [release info]', 'HDR [mediainfo]'),
        (None, 'path_does_exist', '', 'HDR [release info]', 'HDR [release info]'),
        (None, 'path_does_exist', '', '', ''),
    ),
    ids=lambda v: str(v),
)
def test_hdr_format_getter(_hdr_format, path_exists, hdr_from_mediainfo, hdr_from_release_info, exp_value, mocker):
    rn = ReleaseName('path/to/something')

    mocker.patch('os.path.exists', Mock(return_value=True if path_exists == 'path_does_exist' else False))
    mocker.patch('upsies.utils.video.hdr_format', Mock(return_value=hdr_from_mediainfo))
    mocker.patch.object(rn, '_info', {
        'hdr_format': hdr_from_release_info,
    })
    if _hdr_format is not None:
        mocker.patch.object(rn, '_hdr_format', _hdr_format, create=True)

    assert rn.hdr_format == exp_value

@pytest.mark.parametrize(
    argnames='value, exp_result',
    argvalues=(
        (None, ''),
        ('', ''),
        ('DV', 'DV'),
        ('HDR', 'HDR'),
        ('foo', ValueError("Unknown HDR format: 'foo'")),
    ),
    ids=lambda v: str(v),
)
def test_hdr_format_setter(value, exp_result, mocker):
    rn = ReleaseName('path/to/something')

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            rn.hdr_format = value
    else:
        rn.hdr_format = value
        assert rn.hdr_format == exp_result

def test_hdr_format_is_translated(mocker):
    mocker.patch('upsies.utils.release.ReleaseInfo', Mock(return_value={'hdr_format': 'HDR10'}))
    translation = {
        'hdr_format': {
            re.compile(r'(?i:hdr)'): r'ABC',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    assert rn.hdr_format == 'ABC10'
    rn.hdr_format = 'DV'
    assert rn.hdr_format == 'DV'
    rn.hdr_format = 'HDR10+'
    assert rn.hdr_format == 'ABC10+'


@patch('upsies.utils.release.ReleaseInfo')
@patch('upsies.utils.video.video_format')
def test_video_format_getter_prefers_mediainfo(video_format_mock, ReleaseInfo_mock):
    video_format_mock.return_value = 'x264'
    ReleaseInfo_mock.return_value = {'video_codec': 'x265'}
    assert ReleaseName('path/to/something').video_format == 'x264'
    assert video_format_mock.call_args_list == [call('path/to/something', default=None)]

@patch('upsies.utils.release.ReleaseInfo')
@patch('upsies.utils.video.video_format')
def test_video_format_getter_defaults_to_guess(video_format_mock, ReleaseInfo_mock):
    video_format_mock.return_value = None
    ReleaseInfo_mock.return_value = {'video_codec': 'x265'}
    assert ReleaseName('path/to/something').video_format == 'x265'
    assert video_format_mock.call_args_list == [call('path/to/something', default=None)]

@patch('upsies.utils.release.ReleaseInfo')
@patch('upsies.utils.video.video_format')
def test_video_format_getter_defaults_to_placeholder(video_format_mock, ReleaseInfo_mock):
    video_format_mock.return_value = None
    ReleaseInfo_mock.return_value = {}
    assert ReleaseName('path/to/something').video_format == 'UNKNOWN_VIDEO_FORMAT'
    assert video_format_mock.call_args_list == [call('path/to/something', default=None)]

@patch('upsies.utils.release.ReleaseInfo')
def test_video_format_setter(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {}
    rn = ReleaseName('path/to/something')
    assert rn.video_format == 'UNKNOWN_VIDEO_FORMAT'
    rn.video_format = 'x265'
    assert rn.video_format == 'x265'
    rn.video_format = ''
    assert rn.video_format == 'UNKNOWN_VIDEO_FORMAT'
    rn.video_format = 264
    assert rn.video_format == '264'

@patch('upsies.utils.release.ReleaseInfo')
@patch('upsies.utils.video.video_format')
def test_video_format_is_translated(video_format_mock, ReleaseInfo_mock):
    video_format_mock.return_value = None
    ReleaseInfo_mock.return_value = {'video_codec': 'x264'}
    translation = {
        'video_format': {
            re.compile(r'x(\d+)'): r'H.\1',
            re.compile(r'H\.?'): r'Harry',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    assert rn.video_format == 'Harry264'
    rn.video_format = 'asdf'
    assert rn.video_format == 'asdf'
    rn.video_format = 'H265'
    assert rn.video_format == 'Harry265'


@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_group_getter(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'group': 'ASDF'}
    assert ReleaseName('path/to/something').group == 'ASDF'
    ReleaseInfo_mock.return_value = {'release_group': ''}
    assert ReleaseName('path/to/something').group == 'NOGROUP'
    ReleaseInfo_mock.return_value = {}
    assert ReleaseName('path/to/something').group == 'NOGROUP'

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_group_setter(ReleaseInfo_mock):
    rn = ReleaseName('path/to/something')
    assert rn.group == 'NOGROUP'
    rn.group = 'ASDF'
    assert rn.group == 'ASDF'
    rn.group = ''
    assert rn.group == 'NOGROUP'
    rn.group = 123
    assert rn.group == '123'

@patch('upsies.utils.release.ReleaseInfo', new_callable=lambda: Mock(return_value={}))
def test_group_is_translated(ReleaseInfo_mock):
    ReleaseInfo_mock.return_value = {'group': 'A-SDF'}
    translation = {
        'group': {
            re.compile(r'-'): r'_',
        },
    }
    rn = ReleaseName('path/to/something', translate=translation)
    assert rn.group == 'A_SDF'
    rn.group = 'F-o-o'
    assert rn.group == 'F_o_o'


@pytest.mark.parametrize(
    argnames='path_exists, has_commentary, release_info, exp_value',
    argvalues=(
        ('path_exists', True, {'has_commentary': True}, True),
        ('path_exists', True, {'has_commentary': False}, True),

        ('path_exists', False, {'has_commentary': True}, False),
        ('path_exists', False, {'has_commentary': False}, False),

        ('path_exists', None, {'has_commentary': True}, False),
        ('path_exists', None, {'has_commentary': False}, False),

        ('path_does_not_exist', True, {'has_commentary': True}, True),
        ('path_does_not_exist', True, {'has_commentary': False}, False),

        ('path_does_not_exist', False, {'has_commentary': True}, True),
        ('path_does_not_exist', False, {'has_commentary': False}, False),

        ('path_does_not_exist', None, {'has_commentary': True}, True),
        ('path_does_not_exist', None, {'has_commentary': False}, False),
    ),
    ids=lambda v: str(v),
)
def test_has_commentary(path_exists, has_commentary, release_info, exp_value, mocker):
    mocker.patch(
        'upsies.utils.video.has_commentary',
        Mock(return_value=has_commentary),
    )
    mocker.patch(
        'upsies.utils.release.ReleaseInfo',
        Mock(return_value=release_info),
    )
    mocker.patch(
        'os.path.exists',
        Mock(return_value=True if path_exists == 'path_exists' else False),
    )
    rn = ReleaseName('path/to/something')
    assert rn.has_commentary is exp_value
    rn.has_commentary = ''
    assert rn.has_commentary is False
    rn.has_commentary = 1
    assert rn.has_commentary is True
    rn.has_commentary = None
    assert rn.has_commentary is exp_value


@pytest.mark.parametrize(
    argnames='path_exists, has_dual_audio, release_info, exp_value',
    argvalues=(
        ('path_exists', True, {'edition': ['Dual Audio']}, True),
        ('path_exists', True, {'edition': []}, True),

        ('path_exists', False, {'edition': ['Dual Audio']}, False),
        ('path_exists', False, {'edition': []}, False),

        ('path_exists', None, {'edition': ['Dual Audio']}, False),
        ('path_exists', None, {'edition': []}, False),

        ('path_does_not_exist', True, {'edition': ['Dual Audio']}, True),
        ('path_does_not_exist', True, {'edition': []}, False),

        ('path_does_not_exist', False, {'edition': ['Dual Audio']}, True),
        ('path_does_not_exist', False, {'edition': []}, False),

        ('path_does_not_exist', None, {'edition': ['Dual Audio']}, True),
        ('path_does_not_exist', None, {'edition': []}, False),
    ),
    ids=lambda v: str(v),
)
def test_has_dual_audio(path_exists, has_dual_audio, release_info, exp_value, mocker):
    mocker.patch(
        'upsies.utils.video.has_dual_audio',
        Mock(return_value=has_dual_audio),
    )
    mocker.patch(
        'upsies.utils.release.ReleaseInfo',
        Mock(return_value=release_info),
    )
    mocker.patch(
        'os.path.exists',
        Mock(return_value=True if path_exists == 'path_exists' else False),
    )
    rn = ReleaseName('path/to/something')
    assert rn.has_dual_audio is exp_value
    rn.has_dual_audio = ''
    assert rn.has_dual_audio is False
    rn.has_dual_audio = 1
    assert rn.has_dual_audio is True
    rn.has_dual_audio = None
    assert rn.has_dual_audio is exp_value


@pytest.mark.parametrize('release_type', tuple(ReleaseType), ids=lambda v: repr(v))
def test_is_complete_handles_all_ReleaseTypes(release_type):
    rn = ReleaseName('path/to/something')
    rn.type = release_type
    assert rn.is_complete is False

@pytest.mark.parametrize(
    argnames='release_name, release_type, needed_attrs',
    argvalues=(
        ('The Foo 1998 1080p BluRay DTS x264-ASDF', ReleaseType.movie, ReleaseName._needed_attrs[ReleaseType.movie]),
        ('The Foo S03 1080p BluRay DTS x264-ASDF', ReleaseType.season, ReleaseName._needed_attrs[ReleaseType.season]),
        ('The Foo S03E04 1080p BluRay DTS x264-ASDF', ReleaseType.episode, ReleaseName._needed_attrs[ReleaseType.episode]),
    ),
    ids=lambda v: str(v),
)
def test_is_complete(release_name, release_type, needed_attrs):
    rn = ReleaseName(release_name)
    rn.type = release_type
    assert rn.is_complete is True
    for attr in needed_attrs:
        value = getattr(rn, attr)
        setattr(rn, attr, '')
        assert rn.type is release_type
        assert rn.is_complete is False
        setattr(rn, attr, value)
        assert rn.is_complete is True


@pytest.mark.parametrize(
    argnames='source, frame_rate, exp_dvd_encoding',
    argvalues=(
        ('NonDVD', 25, None),
        ('DVD', 0, None),
        ('DVD', 25, 'PAL'),
        ('DVD', 30, 'NTSC'),
        ('DVD5', 0, None),
        ('DVD5', 25, 'PAL'),
        ('DVD5', 30, 'NTSC'),
        ('DVD9', 0, None),
        ('DVD9', 25, 'PAL'),
        ('DVD9', 30, 'NTSC'),
        ('DVDRip', 0, None),
        ('DVDRip', 25, 'PAL'),
        ('DVDRip', 30, 'NTSC'),
    ),
    ids=lambda v: str(v),
)
def test_dvd_encoding(source, frame_rate, exp_dvd_encoding, mocker):
    release = 'path/to/release'
    rn = ReleaseName(release)
    mocker.patch.object(type(rn), 'source', PropertyMock(return_value=source))
    mocker.patch('upsies.utils.video.frame_rate', return_value=frame_rate)
    dvd_encoding = rn.dvd_encoding
    assert dvd_encoding == exp_dvd_encoding


@pytest.mark.parametrize('callback', (None, Mock()), ids=['without callback', 'with callback'])
@pytest.mark.asyncio
async def test_fetch_info(callback, mocker):
    rn = ReleaseName('Foo 2000 1080p BluRay DTS x264-ASDF')

    mocks = AsyncMock()
    mocker.patch.object(rn, '_update_attributes', mocks._update_attributes)
    mocker.patch.object(rn, '_update_type', mocks._update_type)
    mocker.patch.object(rn, '_update_country', mocks._update_country)
    mocker.patch.object(rn, '_update_year_country_required', mocks._update_year_country_required)

    webdb = Mock()
    webdb_id = 'mock id'

    fetch_info_kwargs = {'webdb': webdb, 'webdb_id': webdb_id, 'callback': callback}
    return_value = await rn.fetch_info(**fetch_info_kwargs)
    assert return_value is rn

    assert mocks.mock_calls == [
        call._update_attributes(webdb, webdb_id),
        call._update_type(webdb, webdb_id),
        call._update_country(webdb, webdb_id),
        call._update_year_country_required(webdb, webdb_id),
    ]
    if callback:
        assert callback.call_args_list == [call(rn)]


@pytest.mark.parametrize('no_value', ('', None))
@pytest.mark.parametrize(
    argnames='info',
    argvalues=(
        {'title_original': 'Le Foo', 'title_english': 'The Foo', 'year': '2000'},
        {'title_original': '', 'title_english': '', 'year': ''},

        {'title_original': 'Le Foo', 'title_english': 'The Foo', 'year': ''},
        {'title_original': 'Le Foo', 'title_english': '', 'year': '2000'},
        {'title_original': '', 'title_english': 'The Foo', 'year': '2000'},

        {'title_original': 'Le Foo', 'title_english': '', 'year': ''},
        {'title_original': '', 'title_english': 'The Foo', 'year': ''},
        {'title_original': '', 'title_english': '', 'year': '2000'},
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_update_attributes(info, no_value, mocker):
    rn = ReleaseName('path/to/something')
    rn.title = 'Guessed Title'
    rn.title_aka = 'Guessed AKA'
    rn.year = '1999'

    webdb = Mock(gather=AsyncMock(return_value={
        k: v if v else no_value
        for k, v in info.items()
    }))
    webdb_id = 'mock id'

    await rn._update_attributes(webdb, webdb_id)

    assert rn.title == (info['title_original'] or 'Guessed Title')
    assert rn.title_aka == (info['title_english'] or 'Guessed AKA')
    assert rn.year == (info['year'] or '1999')

    assert webdb.gather.call_args_list == [
        call(webdb_id, 'title_english', 'title_original', 'year'),
    ]


@pytest.mark.parametrize(
    argnames='guessed_type, webdb_type, exp_type',
    argvalues=(
        (ReleaseType.movie, ReleaseType.movie, ReleaseType.movie),
        (ReleaseType.movie, ReleaseType.season, ReleaseType.season),
        (ReleaseType.movie, ReleaseType.episode, ReleaseType.episode),
        (ReleaseType.movie, ReleaseType.unknown, ReleaseType.movie),
        (ReleaseType.season, ReleaseType.movie, ReleaseType.movie),
        (ReleaseType.season, ReleaseType.season, ReleaseType.season),
        (ReleaseType.season, ReleaseType.episode, ReleaseType.episode),
        (ReleaseType.season, ReleaseType.unknown, ReleaseType.season),
        # For episodes, we trust guessed type more than IMDb et al
        (ReleaseType.episode, ReleaseType.movie, ReleaseType.episode),
        (ReleaseType.episode, ReleaseType.season, ReleaseType.episode),
        (ReleaseType.episode, ReleaseType.episode, ReleaseType.episode),
        (ReleaseType.episode, ReleaseType.unknown, ReleaseType.episode),
    ),
)
@pytest.mark.asyncio
async def test_update_type(guessed_type, webdb_type, exp_type, mocker):
    webdb = Mock(type=AsyncMock(return_value=webdb_type))
    webdb_id = 'mock id'

    rn = ReleaseName('path/to/something')
    mocker.patch.object(rn, '_info', {'type': guessed_type})

    assert rn.type == guessed_type
    await rn._update_type(webdb, webdb_id)
    assert rn.type == exp_type
    assert webdb.type.call_args_list == [call(webdb_id)]


@pytest.mark.parametrize(
    argnames='guessed_country, webdb_countries, exp_country',
    argvalues=(
        ('DK', NotImplementedError('This is not implemented'), 'DK'),
        ('DK', [], 'DK'),
        ('DK', ['Sweden', 'Finland'], 'SE'),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_update_country(guessed_country, webdb_countries, exp_country, mocker):
    if isinstance(webdb_countries, Exception):
        webdb = Mock(countries=AsyncMock(side_effect=webdb_countries))
    else:
        webdb = Mock(countries=AsyncMock(return_value=webdb_countries))
    webdb_id = 'mock id'

    rn = ReleaseName('path/to/something')
    mocker.patch.object(rn, '_info', {'country': guessed_country})

    assert rn.country == guessed_country
    await rn._update_country(webdb, webdb_id)
    assert rn.country == exp_country
    assert webdb.countries.call_args_list == [call(webdb_id)]


@pytest.mark.parametrize('release_type', (ReleaseType.movie, ReleaseType.season, ReleaseType.episode))
@pytest.mark.parametrize(
    argnames='attributes, results, exp_year_required, exp_country_required',
    argvalues=(
        # Series: unique name, irrelevant year, irrelevant countries
        (
            {'title': 'The Foo', 'year': None, 'countries': []},
            (
                ('The Foo', '2000', ['US']),
                ('The Foe', '2000', ['US']),
                ('The Bar', '2000', ['US']),
                ('The Bar 2', None, ['US']),
                ('The Baz 2', '2000', None),
            ),
            None, None,
        ),

        # Series: common name, unique year, irrelevant countries
        (
            {'title': 'The Foo', 'year': '2001', 'countries': []},
            (
                ('The FOO', '2000', ['US']),
                ('The Föö', '2001', ['US']),
                ('The Fóò', '2002', ['US']),
                ('The Bar', '2000', ['US']),
                ('The Baz', '2000', ['US']),
                ('The Bar 2', None, ['US']),
                ('The Baz 2', '2000', None),
            ),
            True, None,
        ),

        # Series: common name, common year, unique countries
        (
            {'title': 'The Foo', 'year': '2002', 'countries': ['UK']},
            (
                ('The FOO', '2002', ['US']),
                ('The Föö', '2002', ['UK']),
                ('The Fóò', '2002', ['US']),
                ('The Bar', '2002', ['US']),
                ('The Baz', '2002', ['US']),
                ('The Bar 2', None, ['US']),
                ('The Baz 2', '2002', None),
            ),
            None, True,
        ),

        # Series: common name, common year, common countries
        (
            {'title': 'The Foo', 'year': '2003', 'countries': ['UK']},
            (
                ('The FOO', '2003', ['US']),
                ('The Föö', '2003', ['UK']),
                ('The Fóò', '2003', ['UK']),
                ('The Bar', '2003', ['US']),
                ('The Bar 2', None, ['US']),
                ('The Baz 2', '2003', None),
            ),
            True, True,
        ),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_update_year_country_required(release_type, attributes, results, exp_year_required, exp_country_required, mocker):
    from types import SimpleNamespace
    mocked_results = [
        SimpleNamespace(title=r[0], year=r[1], countries=AsyncMock(return_value=r[2]))
        for r in results
    ]
    webdb = Mock(
        search=AsyncMock(return_value=mocked_results),
        year=AsyncMock(return_value=attributes['year']),
        countries=AsyncMock(return_value=attributes['countries']),
    )
    webdb_id = 'tt123'

    rn = ReleaseName('path/to/something')
    mocker.patch.object(type(rn), 'type', PropertyMock(return_value=release_type))
    mocker.patch.object(type(rn), 'title', PropertyMock(return_value=attributes['title']))

    orig_year_required = rn.year_required
    orig_country_required = rn.country_required

    if release_type is ReleaseType.movie:
        exp_year_required = orig_year_required
        exp_country_required = orig_country_required
    else:
        if exp_year_required is None:
            exp_year_required = orig_year_required
        if exp_country_required is None:
            exp_country_required = orig_country_required

    await rn._update_year_country_required(webdb, webdb_id)

    assert rn.year_required is exp_year_required
    assert rn.country_required is exp_country_required

    if release_type is ReleaseType.movie:
        assert webdb.mock_calls == []
    else:
        exp_query = webdbs.Query(title='The Foo', type=ReleaseType.series)
        assert webdb.mock_calls == [
            call.search(exp_query),
            call.year(webdb_id),
            call.countries(webdb_id),
        ]


TITLE_WITH_AKA_AND_YEAR = {
    'Foo: Bär... AKA B-a-z & Quux 2003': {
        ' ': 'Foo: Bär... AKA B-a-z & Quux 2003',
        '.': 'Foo.Bär.AKA.B-a-z.and.Quux.2003',
    },
}

@pytest.mark.parametrize('title_with_aka_and_year', TITLE_WITH_AKA_AND_YEAR.keys())
@pytest.mark.parametrize(
    argnames='release_type, exp_episodes_included',
    argvalues=(
        (ReleaseType.movie, False),
        (ReleaseType.season, True),
        (ReleaseType.episode, True),
    ),
)
@pytest.mark.parametrize(
    argnames='resolution, dvd_encoding, exp_resolution',
    argvalues=(
        ('123p', None, '123p'),
        ('123p', 'NTSC-PAL', 'NTSC-PAL'),
    ),
)
@pytest.mark.parametrize(
    argnames='separator_attribute, separator_argument, exp_separator',
    argvalues=(
        (' ', None, ' '),
        (' ', '.', '.'),
        ('.', None, '.'),
        ('.', ' ', ' '),
    ),
)
@pytest.mark.parametrize(
    argnames='attributes, exp_release_name',
    argvalues=(
        (
            {'service': '', 'edition': (), 'audio_format': '', 'audio_channels': ''},
            '{title_with_aka_and_year_and_episodes}{s}{resolution}{s}{source}{s}{video_format}-{group}',
        ),
        (
            {'service': 'ABC', 'edition': (), 'audio_format': '', 'audio_channels': ''},
            '{title_with_aka_and_year_and_episodes}{s}{resolution}{s}ABC{s}{source}{s}{video_format}-{group}',
        ),
        (
            {'service': '', 'edition': ('DC',), 'audio_format': 'DTS', 'audio_channels': ''},
            '{title_with_aka_and_year_and_episodes}{s}DC{s}{resolution}{s}{source}{s}DTS{s}{video_format}-{group}',
        ),
        (
            {'service': '', 'edition': ('DC', 'FE'), 'audio_format': 'DTS', 'audio_channels': ''},
            '{title_with_aka_and_year_and_episodes}{s}DC{s}FE{s}{resolution}{s}{source}{s}DTS{s}{video_format}-{group}',
        ),
        (
            {'service': '', 'edition': (), 'audio_format': 'DTS', 'audio_channels': '3.4'},
            '{title_with_aka_and_year_and_episodes}{s}{resolution}{s}{source}{s}DTS{s}3.4{s}{video_format}-{group}',
        ),
        (
            {'service': 'ABC', 'edition': ('DC',), 'audio_format': 'DTS', 'audio_channels': '3.4'},
            '{title_with_aka_and_year_and_episodes}{s}DC{s}{resolution}{s}ABC{s}{source}{s}DTS{s}3.4{s}{video_format}-{group}',
        ),
        (
            {'service': '', 'edition': (), 'audio_format': '', 'audio_channels': '', 'hdr_format': 'HDR10',},
            '{title_with_aka_and_year_and_episodes}{s}{resolution}{s}{source}{s}HDR10{s}{video_format}-{group}',
        ),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_format(
        title_with_aka_and_year,
        release_type, exp_episodes_included,
        resolution, dvd_encoding, exp_resolution,
        separator_attribute, separator_argument, exp_separator,
        attributes, exp_release_name,
        mocker,
):
    rn = ReleaseName('irrelevant/path/to/content')

    static_attributes = {
        'episodes': 'S__E__',
        'source': 'DLRip',
        'video_format': 'VC',
        'group': 'ASDF',
    }
    for name, value in static_attributes.items():
        mocker.patch.object(type(rn), name, PropertyMock(return_value=value))
    for name, value in attributes.items():
        mocker.patch.object(type(rn), name, PropertyMock(return_value=value))

    mocker.patch.object(type(rn), 'type', PropertyMock(return_value=release_type))
    mocker.patch.object(type(rn), 'dvd_encoding', PropertyMock(return_value=dvd_encoding))
    mocker.patch.object(type(rn), 'resolution', PropertyMock(return_value=resolution))
    mocker.patch.object(type(rn), 'separator', PropertyMock(return_value=separator_attribute))
    mocker.patch.object(type(rn), 'title_with_aka_and_year', PropertyMock(return_value=title_with_aka_and_year))
    exp_title_with_aka_and_year_and_episodes = TITLE_WITH_AKA_AND_YEAR[title_with_aka_and_year][exp_separator]
    if exp_episodes_included:
        exp_title_with_aka_and_year_and_episodes += exp_separator + static_attributes['episodes']

    release_name = rn.format(separator=separator_argument)
    exp_release_name_resolved = exp_release_name.format(
        s=exp_separator,
        title_with_aka_and_year_and_episodes=exp_title_with_aka_and_year_and_episodes,
        resolution=exp_resolution,
        **static_attributes,
    )
    assert release_name == exp_release_name_resolved



from types import SimpleNamespace
from unittest.mock import DEFAULT, AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import __homepage__, __project_name__, utils
from upsies.trackers import bhd


@pytest.fixture
def tracker():
    tracker = Mock()
    tracker.name = 'bhd'
    return tracker


@pytest.fixture
def imghost():
    class MockImageHost(utils.imghosts.base.ImageHostBase):
        name = 'mock image host'
        default_config = {}
        _upload_image = AsyncMock()

    return MockImageHost()


@pytest.fixture
def bhd_tracker_jobs(imghost, tracker, tmp_path, mocker):
    content_path = tmp_path / 'Foo 2000 1080p BluRay x264-ASDF'

    bhd_tracker_jobs = bhd.BhdTrackerJobs(
        content_path=str(content_path),
        tracker=tracker,
        image_hosts=(imghost,),
        btclient=Mock(),
        torrent_destination=str(tmp_path / 'destination'),
        common_job_args={
            'home_directory': str(tmp_path / 'home_directory'),
            'ignore_cache': True,
        },
        options=None,
    )

    return bhd_tracker_jobs


@pytest.fixture
def mock_job_attributes(mocker):
    def mock_job_attributes(bhd_tracker_jobs):
        job_attrs = (
            # Background jobs
            'create_torrent_job',
            'mediainfo_job',
            'screenshots_job',
            'upload_screenshots_job',

            # Interactive jobs
            'tmdb_job',
            'imdb_job',
            'release_name_job',
            'category_job',
            'type_job',
            'source_job',
            'description_job',
            'scene_check_job',
            'tags_job',
        )
        for job_attr in job_attrs:
            mocker.patch.object(type(bhd_tracker_jobs), job_attr, PropertyMock(return_value=Mock(attr=job_attr)))

    return mock_job_attributes


def test_jobs_before_upload_items(bhd_tracker_jobs, mock_job_attributes, mocker):
    mock_job_attributes(bhd_tracker_jobs)

    print(bhd_tracker_jobs.jobs_before_upload)
    assert tuple(job.attr for job in bhd_tracker_jobs.jobs_before_upload) == (
        # Background jobs
        'create_torrent_job',
        'mediainfo_job',
        'screenshots_job',
        'upload_screenshots_job',

        # Interactive jobs
        'tmdb_job',
        'imdb_job',
        'release_name_job',
        'category_job',
        'type_job',
        'source_job',
        'description_job',
        'scene_check_job',
        'tags_job',
    )


def test_isolated_jobs__only_description(bhd_tracker_jobs, mock_job_attributes, mocker):
    mock_job_attributes(bhd_tracker_jobs)
    mocker.patch.object(type(bhd_tracker_jobs), 'options', PropertyMock(return_value={'only_description': True}))
    mocker.patch.object(bhd_tracker_jobs, 'get_job_and_dependencies')
    assert bhd_tracker_jobs.isolated_jobs is bhd_tracker_jobs.get_job_and_dependencies.return_value
    assert bhd_tracker_jobs.get_job_and_dependencies.call_args_list == [
        call(bhd_tracker_jobs.description_job, bhd_tracker_jobs.screenshots_job)
    ]

def test_isolated_jobs__only_title(bhd_tracker_jobs, mock_job_attributes, mocker):
    mock_job_attributes(bhd_tracker_jobs)
    mocker.patch.object(type(bhd_tracker_jobs), 'options', PropertyMock(return_value={'only_title': True}))
    mocker.patch.object(bhd_tracker_jobs, 'get_job_and_dependencies')
    assert bhd_tracker_jobs.isolated_jobs is bhd_tracker_jobs.get_job_and_dependencies.return_value
    assert bhd_tracker_jobs.get_job_and_dependencies.call_args_list == [
        call(bhd_tracker_jobs.release_name_job, bhd_tracker_jobs.imdb_job)
    ]

def test_isolated_jobs__no_isolated_jobs(bhd_tracker_jobs, mock_job_attributes, mocker):
    mock_job_attributes(bhd_tracker_jobs)
    mocker.patch.object(type(bhd_tracker_jobs), 'options', PropertyMock(return_value={}))
    mocker.patch.object(bhd_tracker_jobs, 'get_job_and_dependencies')
    assert bhd_tracker_jobs.isolated_jobs == ()
    assert bhd_tracker_jobs.get_job_and_dependencies.call_args_list == []


def test_tmdb_job(bhd_tracker_jobs, mocker):
    mocker.patch('upsies.jobs.webdb.WebDbSearchJob')
    assert bhd_tracker_jobs.tmdb_job.no_id_ok is True


def test_category_job(bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), 'release_name_job', PropertyMock())
    ChoiceJob_mock = mocker.patch('upsies.jobs.dialog.ChoiceJob')
    mocker.patch.object(bhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(bhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(bhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert bhd_tracker_jobs.category_job is ChoiceJob_mock.return_value
    assert ChoiceJob_mock.call_args_list == [call(
        name=bhd_tracker_jobs.get_job_name.return_value,
        label='Category',
        precondition=bhd_tracker_jobs.make_precondition.return_value,
        prejobs=(
            bhd_tracker_jobs.release_name_job,
        ),
        autodetect=bhd_tracker_jobs.autodetect_category,
        options=(
            ('Movie', '1'),
            ('TV', '2'),
        ),
        common_job_arg='common job argument',
    )]
    assert bhd_tracker_jobs.get_job_name.call_args_list == [call('category')]
    assert bhd_tracker_jobs.make_precondition.call_args_list == [call('category_job')]
    assert bhd_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='release_type, exp_return_value',
    argvalues=(
        (utils.release.ReleaseType.movie, 'Movie'),
        (utils.release.ReleaseType.series, 'TV'),
        (utils.release.ReleaseType.episode, 'TV'),
        (None, None),
    ),
    ids=lambda v: str(v),
)
def test_autodetect_category(release_type, exp_return_value, bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), 'release_name', PropertyMock(return_value=Mock(
        type=release_type,
    )))
    return_value = bhd_tracker_jobs.autodetect_category('_')
    assert return_value == exp_return_value


def test_type_job(bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), 'release_name_job', PropertyMock())
    ChoiceJob_mock = mocker.patch('upsies.jobs.dialog.ChoiceJob')
    mocker.patch.object(bhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(bhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(bhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert bhd_tracker_jobs.type_job is ChoiceJob_mock.return_value
    assert ChoiceJob_mock.call_args_list == [call(
        name=bhd_tracker_jobs.get_job_name.return_value,
        label='Type',
        precondition=bhd_tracker_jobs.make_precondition.return_value,
        prejobs=(
            bhd_tracker_jobs.release_name_job,
        ),
        autodetect_and_finish=bhd_tracker_jobs.autodetect_type,
        options=(
            ('UHD 100', 'UHD 100'),
            ('UHD 66', 'UHD 66'),
            ('UHD 50', 'UHD 50'),
            ('UHD Remux', 'UHD Remux'),
            ('BD 50', 'BD 50'),
            ('BD 25', 'BD 25'),
            ('BD Remux', 'BD Remux'),
            ('2160p', '2160p'),
            ('1080p', '1080p'),
            ('1080i', '1080i'),
            ('720p', '720p'),
            ('576p', '576p'),
            ('540p', '540p'),
            ('DVD 9', 'DVD 9'),
            ('DVD 5', 'DVD 5'),
            ('DVD Remux', 'DVD Remux'),
            ('480p', '480p'),
            ('Other', 'Other'),
        ),
        focused='Other',
        common_job_arg='common job argument',
    )]
    assert bhd_tracker_jobs.get_job_name.call_args_list == [call('type')]
    assert bhd_tracker_jobs.make_precondition.call_args_list == [call('type_job')]
    assert bhd_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='resolution, source, exp_return_value',
    argvalues=(
        ('', 'DVD9', 'DVD 9'),
        ('', 'DVD5', 'DVD 5'),
        ('', 'DVD Remux', 'DVD Remux'),
        ('2160p', '', '2160p'),
        ('1080p', '', '1080p'),
        ('1080i', '', '1080i'),
        ('720p', '', '720p'),
        ('576p', '', '576p'),
        ('540p', '', '540p'),
        ('480p', '', '480p'),
        ('123p', '', None),
    ),
    ids=lambda v: str(v),
)
def test_autodetect_type(resolution, source, exp_return_value, bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), 'release_name', PropertyMock(return_value=Mock(
        resolution=resolution,
        source=source,
    )))
    return_value = bhd_tracker_jobs.autodetect_type('_')
    assert return_value == exp_return_value


def test_source_job(bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), 'release_name_job', PropertyMock())
    ChoiceJob_mock = mocker.patch('upsies.jobs.dialog.ChoiceJob')
    mocker.patch.object(bhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(bhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(bhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert bhd_tracker_jobs.source_job is ChoiceJob_mock.return_value
    assert ChoiceJob_mock.call_args_list == [call(
        name=bhd_tracker_jobs.get_job_name.return_value,
        label='Source',
        precondition=bhd_tracker_jobs.make_precondition.return_value,
        prejobs=(
            bhd_tracker_jobs.release_name_job,
        ),
        autodetect_and_finish=bhd_tracker_jobs.autodetect_source,
        options=(
            ('Blu-ray', 'Blu-ray'),
            ('HD-DVD', 'HD-DVD'),
            ('WEB', 'WEB'),
            ('HDTV', 'HDTV'),
            ('DVD', 'DVD'),
        ),
        common_job_arg='common job argument',
    )]
    assert bhd_tracker_jobs.get_job_name.call_args_list == [call('source')]
    assert bhd_tracker_jobs.make_precondition.call_args_list == [call('source_job')]
    assert bhd_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='source, exp_return_value',
    argvalues=(
        ('BluRay', 'Blu-ray'),
        ('BluRay Remux', 'Blu-ray'),
        ('WEB-DL', 'WEB'),
        ('WEBRip', 'WEB'),
        ('WEB', 'WEB'),
        ('DVD9', 'DVD'),
        ('DVD5', 'DVD'),
        ('DVD Remux', 'DVD'),
        ('Foo', None),
    ),
    ids=lambda v: str(v),
)
def test_autodetect_source(source, exp_return_value, bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), 'release_name', PropertyMock(return_value=Mock(
        source=source,
    )))
    return_value = bhd_tracker_jobs.autodetect_source('_')
    assert return_value == exp_return_value


def test_description_job(bhd_tracker_jobs, mocker):
    mocker.patch.object(bhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(bhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(bhd_tracker_jobs, 'generate_description', Mock())
    mocker.patch.object(type(bhd_tracker_jobs), 'upload_screenshots_job', PropertyMock())
    mocker.patch.object(type(bhd_tracker_jobs), 'mediainfo_job', PropertyMock())
    mocker.patch.object(bhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')

    assert bhd_tracker_jobs.description_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=bhd_tracker_jobs.get_job_name.return_value,
        label='Description',
        precondition=bhd_tracker_jobs.make_precondition.return_value,
        prejobs=(
            bhd_tracker_jobs.upload_screenshots_job,
            bhd_tracker_jobs.mediainfo_job,
        ),
        text=bhd_tracker_jobs.generate_description,
        hidden=True,
        finish_on_success=True,
        read_only=True,
        common_job_arg='common job argument',
    )]
    assert bhd_tracker_jobs.get_job_name.call_args_list == [call('description')]
    assert bhd_tracker_jobs.make_precondition.call_args_list == [call('description_job')]
    assert bhd_tracker_jobs.common_job_args.call_args_list == [call(ignore_cache=True)]


def test_image_host_config(bhd_tracker_jobs, mocker):
    assert bhd_tracker_jobs.image_host_config == {
        'common': {'thumb_width': 350},
    }


@pytest.mark.parametrize(
    argnames='screenshot_tags, mediainfo_tags, exp_bbcode',
    argvalues=(
        (
            '<screenshot tags>',
            '',
            (
                '[center]\n<screenshot tags>\n[/center]\n\n'
                f'[align=right][size=1]Shared with [url={__homepage__}]{__project_name__}[/url][/size][/align]'
            ),
        ),
        (
            '<screenshot tags>',
            '<mediainfo tags>',
            (
                '[center]\n<screenshot tags>\n[/center]\n'
                '[center][h3]Mediainfo[/h3][/center]\n'
                '<mediainfo tags>\n\n'
                f'[align=right][size=1]Shared with [url={__homepage__}]{__project_name__}[/url][/size][/align]'
            ),
        ),
    ),
)
def test_generate_description(screenshot_tags, mediainfo_tags, exp_bbcode, bhd_tracker_jobs, mocker):
    mocker.patch.object(bhd_tracker_jobs, '_generate_description_screenshots', return_value=screenshot_tags)
    mocker.patch.object(bhd_tracker_jobs, '_generate_description_mediainfos', return_value=mediainfo_tags)
    bbcode = bhd_tracker_jobs.generate_description()
    assert bbcode == exp_bbcode


def test__generate_description_screenshots(bhd_tracker_jobs, mocker):
    mocker.patch.object(bhd_tracker_jobs, 'make_screenshots_grid')
    mocker.patch.object(bhd_tracker_jobs, 'upload_screenshots_job', return_value=Mock(
        is_finished=True,
    ))

    return_value = bhd_tracker_jobs._generate_description_screenshots()
    assert return_value == bhd_tracker_jobs.make_screenshots_grid.return_value
    assert bhd_tracker_jobs.make_screenshots_grid.call_args_list == [
        call(
            screenshots=bhd_tracker_jobs.upload_screenshots_job.uploaded_images,
            columns=2,
            horizontal_spacer='   ',
            vertical_spacer='\n\n',
        ),
    ]


@pytest.mark.parametrize(
    argnames='content_path, description_mediainfos, exp_bbcode',
    argvalues=(
        (
            '/path/to/content',
            {},
            '',
        ),
        (
            '/path/to/foo.mkv',
            {
                '/path/to/foo.mkv': '<mediainfo for foo.mkv>',
            },
            '[hide=foo.mkv][code]<mediainfo for foo.mkv>[/code][/hide]',
        ),
        (
            '/path/to/content',
            {
                '/path/to/content/foo.mkv': '<mediainfo for foo.mkv>',
            },
            '[hide=content/foo.mkv][code]<mediainfo for foo.mkv>[/code][/hide]',
        ),
        (
            '/path/to/content',
            {
                '/path/to/content/foo.s01e01.mkv': '<mediainfo for foo.s01e01.mkv>',
                '/path/to/content/foo.s01e02.mkv': '<mediainfo for foo.s01e02.mkv>',
                '/path/to/content/foo.s01e03.mkv': '<mediainfo for foo.s01e03.mkv>',
            },
            (
                '[hide=content/foo.s01e01.mkv][code]<mediainfo for foo.s01e01.mkv>[/code][/hide]\n'
                '[hide=content/foo.s01e02.mkv][code]<mediainfo for foo.s01e02.mkv>[/code][/hide]\n'
                '[hide=content/foo.s01e03.mkv][code]<mediainfo for foo.s01e03.mkv>[/code][/hide]'
            ),
        ),
    ),
    ids=lambda v: repr(v),
)
def test__generate_description_mediainfos(content_path, description_mediainfos, exp_bbcode, bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), '_description_mediainfos', PropertyMock(
        return_value=description_mediainfos,
    ))
    mocker.patch.object(type(bhd_tracker_jobs), 'content_path', PropertyMock(
        return_value=content_path,
    ))
    bbcode = bhd_tracker_jobs._generate_description_mediainfos()
    assert bbcode == exp_bbcode


@pytest.mark.parametrize(
    argnames='mediainfos_by_file, exp_description_mediainfos',
    argvalues=(
        pytest.param(
            {
                'path/to/foo.mkv': '<single mediainfo does not go in description>',
            },
            {},
            id='Single mediainfo',
        ),
        pytest.param(
            {
                'path/to/foo/foo.s01e01.mkv': '<mediainfo for foo.s01e01.mkv>',
                'path/to/foo/foo.s01e02.mkv': '<mediainfo for foo.s01e02.mkv>',
                'path/to/foo/foo.s01e03.mkv': '',  # mediainfo may be empty
                'path/to/foo/foo.s01e04.mkv': '<mediainfo for foo.s01e04.mkv>',
            },
            {
                'path/to/foo/foo.s01e01.mkv': '<mediainfo for foo.s01e01.mkv>',
                'path/to/foo/foo.s01e02.mkv': '<mediainfo for foo.s01e02.mkv>',
                'path/to/foo/foo.s01e04.mkv': '<mediainfo for foo.s01e04.mkv>',
            },
            id='Multiple files (season)',
        ),
        pytest.param(
            {
                'path/to/foo/VIDEO_TS/VTS_01_0.IFO': '<mediainfo for VTS_01_0.IFO>',
                'path/to/foo/VIDEO_TS/VTS_01_1.VOB': '<mediainfo for VTS_01_1.VOB>',
            },
            {
                'path/to/foo/VIDEO_TS/VTS_01_0.IFO': '<mediainfo for VTS_01_0.IFO>',
                'path/to/foo/VIDEO_TS/VTS_01_1.VOB': '<mediainfo for VTS_01_1.VOB>',
            },
            id='Multiple mediainfos (DVD)',
        ),
    ),
)
def test__description_mediainfos(mediainfos_by_file, exp_description_mediainfos, bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), 'mediainfo_job', PropertyMock(return_value=Mock(
        is_finished=True,
        mediainfos_by_file=mediainfos_by_file,
    )))
    description_mediainfos = bhd_tracker_jobs._description_mediainfos
    assert description_mediainfos == exp_description_mediainfos


def test_mediainfo_from_all_videos(bhd_tracker_jobs, mocker):
    assert bhd_tracker_jobs.mediainfo_from_all_videos is False


def test_screenshots_from_all_videos(bhd_tracker_jobs, mocker):
    assert bhd_tracker_jobs.screenshots_from_all_videos is False


def test_tags_job(bhd_tracker_jobs, mocker):
    mocker.patch.object(bhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(bhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(bhd_tracker_jobs, 'generate_tags', Mock())
    mocker.patch.object(type(bhd_tracker_jobs), 'release_name_job', PropertyMock())
    mocker.patch.object(type(bhd_tracker_jobs), 'scene_check_job', PropertyMock())
    mocker.patch.object(bhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')

    assert bhd_tracker_jobs.tags_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=bhd_tracker_jobs.get_job_name.return_value,
        label='Tags',
        precondition=bhd_tracker_jobs.make_precondition.return_value,
        prejobs=(
            bhd_tracker_jobs.release_name_job,
            bhd_tracker_jobs.scene_check_job,
        ),
        text=bhd_tracker_jobs.generate_tags,
        finish_on_success=True,
        read_only=True,
        common_job_arg='common job argument',
    )]
    assert bhd_tracker_jobs.get_job_name.call_args_list == [call('tags')]
    assert bhd_tracker_jobs.make_precondition.call_args_list == [call('tags_job')]
    assert bhd_tracker_jobs.common_job_args.call_args_list == [call()]


class MockReleaseName(SimpleNamespace):
    def __init__(self, **kwargs):
        self.source = []
        self.has_commentary = False
        self.has_dual_audio = False
        self.edition = []
        for k, v in kwargs.items():
            setattr(self, k, v)

class MockOptions(dict):
    def __init__(self, **kwargs):
        self['personal_rip'] = False
        for k, v in kwargs.items():
            self[k] = v

@pytest.mark.parametrize('is_scene_release', (False, True), ids=('is scene', 'not scene'))
@pytest.mark.parametrize('hybrid, hybrid_position', ((None, None), ('Hybrid', 0), ('Hybrid', -1)), ids=lambda v: str(v))
@pytest.mark.parametrize(
    argnames='release_name, options, exp_tags',
    argvalues=(
        (MockReleaseName(source=['WEBRip']), MockOptions(), ('WEBRip',)),
        (MockReleaseName(source=['WEB-DL']), MockOptions(), ['WEBDL']),
        (MockReleaseName(has_commentary=True), MockOptions(), ['Commentary']),
        (MockReleaseName(has_dual_audio=True), MockOptions(), ['DualAudio']),
        (MockReleaseName(edition='Open Matte'), MockOptions(), ['OpenMatte']),
        (MockReleaseName(edition='2in1'), MockOptions(), ['2in1']),
        (MockReleaseName(edition='4k Remastered'), MockOptions(), ['4kRemaster']),
        (MockReleaseName(), MockOptions(personal_rip=True), ['Personal']),
        (MockReleaseName(), MockOptions(), []),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_generate_tags(release_name, options, hybrid, hybrid_position, is_scene_release, exp_tags, bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), 'release_name_job', PropertyMock(return_value=Mock(
        is_finished=True,
    )))
    mocker.patch.object(type(bhd_tracker_jobs), 'scene_check_job', PropertyMock(return_value=Mock(
        is_finished=True,
    )))
    mocker.patch.object(type(bhd_tracker_jobs), 'release_name', PropertyMock(return_value=release_name))
    mocker.patch.object(type(bhd_tracker_jobs), 'options', options)

    def get_job_attribute(job, attribute):
        if job is bhd_tracker_jobs.scene_check_job:
            return is_scene_release
        else:
            return DEFAULT

    mocker.patch.object(bhd_tracker_jobs, 'get_job_attribute', side_effect=get_job_attribute, return_value=None)

    exp_tags = list(exp_tags)

    if hybrid:
        bhd_tracker_jobs.release_name.source.insert(hybrid_position, 'Hybrid')
        exp_tags.append('Hybrid')

    if is_scene_release:
        exp_tags.append('Scene')

    tags = sorted((await bhd_tracker_jobs.generate_tags()).split())
    assert tags == sorted(exp_tags)

    assert bhd_tracker_jobs.get_job_attribute.call_args_list == [call(bhd_tracker_jobs.scene_check_job, 'is_scene_release')]


@pytest.mark.parametrize('draft, exp_live', ((True, '0'), (False, '1'),))
@pytest.mark.parametrize('anonymous, exp_anon', ((True, '1'), (False, '0'),))
def test_post_data(anonymous, exp_anon, draft, exp_live, bhd_tracker_jobs, mock_job_attributes, mocker):
    mock_job_attributes(bhd_tracker_jobs)
    mocker.patch.object(bhd_tracker_jobs, 'get_job_output', side_effect=(
        'mock release name',
        'tt00123456',
        'mock description',
        'Some\nMock\nTags',
    ))
    mocker.patch.object(bhd_tracker_jobs, 'get_job_attribute', side_effect=(
        'mock category',
        'mock type',
        'mock source',
    ))
    mocker.patch.object(type(bhd_tracker_jobs), 'options', PropertyMock(return_value={
        'custom_edition': 'mock custom edition',
        'anonymous': anonymous,
        'draft': draft,
    }))
    mocker.patch.object(type(bhd_tracker_jobs), 'post_data_tmdb_id', PropertyMock(return_value='1234'))
    mocker.patch.object(type(bhd_tracker_jobs), 'post_data_edition', PropertyMock(return_value='mock edition'))
    mocker.patch.object(type(bhd_tracker_jobs), 'read_nfo', Mock(return_value='mock nfo'))
    mocker.patch.object(type(bhd_tracker_jobs), 'post_data_pack', PropertyMock(return_value='mock pack'))
    mocker.patch.object(type(bhd_tracker_jobs), 'post_data_sd', PropertyMock(return_value='mock sd'))
    mocker.patch.object(type(bhd_tracker_jobs), 'post_data_special', PropertyMock(return_value='mock special'))

    assert bhd_tracker_jobs.post_data == {
        'name': 'mock release name',
        'category_id': 'mock category',
        'type': 'mock type',
        'source': 'mock source',
        'imdb_id': 'tt00123456',
        'tmdb_id': '1234',
        'description': 'mock description',
        'edition': 'mock edition',
        'custom_edition': 'mock custom edition',
        'tags': 'Some,Mock,Tags',
        'nfo': 'mock nfo',
        'pack': 'mock pack',
        'sd': 'mock sd',
        'special': 'mock special',
        'anon': exp_anon,
        'live': exp_live,
    }
    assert bhd_tracker_jobs.get_job_output.call_args_list == [
        call(bhd_tracker_jobs.release_name_job, slice=0),
        call(bhd_tracker_jobs.imdb_job, slice=0),
        call(bhd_tracker_jobs.description_job, slice=0),
        call(bhd_tracker_jobs.tags_job, slice=0),
    ]
    assert bhd_tracker_jobs.get_job_attribute.call_args_list == [
        call(bhd_tracker_jobs.category_job, 'choice'),
        call(bhd_tracker_jobs.type_job, 'choice'),
        call(bhd_tracker_jobs.source_job, 'choice'),
    ]


@pytest.mark.parametrize(
    argnames='tmdb_job_output, exp_tmdb_id, exp_get_job_output_called',
    argvalues=(
        ((), 0, False),
        (('movie/1234'), '1234', True),
    ),
    ids=lambda v: repr(v),
)
def test_post_data_tmdb_id(tmdb_job_output, exp_tmdb_id, exp_get_job_output_called, bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), 'tmdb_job', PropertyMock(return_value=Mock(
        output=tmdb_job_output,
    )))
    mocker.patch.object(type(bhd_tracker_jobs), 'get_job_output', return_value=tmdb_job_output)
    assert bhd_tracker_jobs.post_data_tmdb_id == exp_tmdb_id
    if exp_get_job_output_called:
        assert bhd_tracker_jobs.get_job_output.call_args_list == [
            call(bhd_tracker_jobs.tmdb_job, slice=0),
        ]
    else:
        assert bhd_tracker_jobs.get_job_output.call_args_list == []


@pytest.mark.parametrize(
    argnames='edition, exp_edition',
    argvalues=(
        ("Collector's Edition", 'Collector'),
        ("Director's Cut", 'Director'),
        ('Extended Cut', 'Extended'),
        ('Limited', 'Limited'),
        ('Special Edition', 'Special'),
        ('Theatrical Cut', 'Theatrical'),
        ('Uncut', 'Uncut'),
        ('Uncensored', 'Uncut'),
        ('Unrated', 'Unrated'),
        ('Super Duper Custom Cut', None),
        ('', None),
    ),
)
def test_post_data_edition(edition, exp_edition, bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), 'release_name', PropertyMock(return_value=Mock(
        edition=edition,
    )))
    assert bhd_tracker_jobs.post_data_edition == exp_edition


@pytest.mark.parametrize(
    argnames='approved_type, exp_pack',
    argvalues=(
        (utils.types.ReleaseType.movie, '0'),
        (utils.types.ReleaseType.season, '1'),
        (utils.types.ReleaseType.episode, '0'),
    ),
)
def test_post_data_pack(approved_type, exp_pack, bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), 'release_name', PropertyMock(return_value=Mock(
        type=approved_type,
    )))
    assert bhd_tracker_jobs.post_data_pack == exp_pack


@pytest.mark.parametrize(
    argnames='resolution, exp_sd',
    argvalues=(
        ('2160p', '0'),
        ('1080p', '0'),
        ('1080i', '0'),
        ('720p', '0'),
        ('576p', '1'),
        ('540p', '1'),
        ('480p', '1'),
        ('asdf', '0'),
        ('', '0'),
    ),
)
def test_post_data_sd(resolution, exp_sd, bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), 'release_name', PropertyMock(return_value=Mock(
        resolution=resolution,
    )))
    assert bhd_tracker_jobs.post_data_sd == exp_sd


@pytest.mark.parametrize(
    argnames='approved_type, options, exp_special',
    argvalues=(
        (utils.types.ReleaseType.movie, {'special': False}, '0'),
        (utils.types.ReleaseType.season, {'special': False}, '0'),
        (utils.types.ReleaseType.episode, {'special': False}, '0'),
        (utils.types.ReleaseType.movie, {'special': True}, '0'),
        (utils.types.ReleaseType.season, {'special': True}, '0'),
        (utils.types.ReleaseType.episode, {'special': True}, '1'),
    ),
)
def test_post_data_special(approved_type, options, exp_special, bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), 'release_name', PropertyMock(return_value=Mock(
        type=approved_type,
    )))
    mocker.patch.object(type(bhd_tracker_jobs), 'options', PropertyMock(return_value=options))
    assert bhd_tracker_jobs.post_data_special == exp_special


def test_mediainfo_filehandle(bhd_tracker_jobs, mocker):
    mocker.patch.object(type(bhd_tracker_jobs), 'mediainfo_job', PropertyMock(return_value=Mock(
        output=('mock mediainfo',),
    )))
    assert bhd_tracker_jobs.mediainfo_filehandle.read() == b'mock mediainfo'



from unittest.mock import AsyncMock, PropertyMock

from upsies import __project_name__, utils
from upsies.trackers import base


def make_MockTracker(**kwargs):
    class MockTracker(base.TrackerBase):
        name = 'asdf'
        label = 'AsdF'
        TrackerJobs = PropertyMock()
        TrackerConfig = PropertyMock()
        login = AsyncMock()
        logout = AsyncMock()
        _login = AsyncMock()
        _logout = AsyncMock()
        is_logged_in = PropertyMock()
        get_announce_url = AsyncMock()
        upload = AsyncMock()

    return MockTracker(**kwargs)


def test_Howto_current_section_and_bump_section():
    howto = base._howto.Howto(tracker_cls=type(make_MockTracker()))
    for i in range(10):
        for _ in range(3):
            assert howto.current_section == i

        assert howto.bump_section == ''

        for _ in range(3):
            assert howto.current_section == i + 1

def test_Howto_introduction():
    howto = base._howto.Howto(tracker_cls=type(make_MockTracker()))
    howto._section = 123
    assert howto.introduction == (
        '123. How To Read This Howto\n'
        '\n'
        '   123.1 Words in ALL_CAPS_AND_WITH_UNDERSCORES are placeholders.\n'
        '   123.2 Everything after "$" is a terminal command.'
    )

def test_Howto_screenshots():
    tracker_cls = type(make_MockTracker())
    howto = base._howto.Howto(tracker_cls=tracker_cls)
    howto._section = 6
    assert howto.screenshots == (
        '6. Screenshots (Optional)\n'
        '\n'
        '   6.1 Specify how many screenshots to make.\n'
        f'       $ {__project_name__} set trackers.{tracker_cls.name}.screenshots NUMBER_OF_SCREENSHOTS\n'
        '\n'
        '   6.2 Specify where to host images.\n'
        f'       $ {__project_name__} set trackers.{tracker_cls.name}.image_host IMAGE_HOST,...\n'
        '       Supported services: dummy, freeimage, imgbb, imgbox, ptpimg\n'
        '\n'
        '   6.3 Configure image hosting service.\n'
        f'       $ {__project_name__} upload-images IMAGE_HOST --help'
    )

def test_Howto_autoseed():
    tracker_cls = type(make_MockTracker())
    howto = base._howto.Howto(tracker_cls=tracker_cls)
    howto._section = 7
    assert howto.autoseed == (
        '7. Add Uploaded Torrents To Client (Optional)\n'
        '\n'
        '   7.1 Specify which client to add uploaded torrents to.\n'
        f'       $ {__project_name__} set trackers.{tracker_cls.name}.add_to CLIENT_NAME\n'
        f'       Supported clients: {", ".join(n for n in utils.btclient.client_names())}\n'
        '\n'
        '   7.2 Specify your client connection.\n'
        f'       $ {__project_name__} set clients.CLIENT_NAME.url URL\n'
        f'       $ {__project_name__} set clients.CLIENT_NAME.username USERNAME\n'
        f'       $ {__project_name__} set clients.CLIENT_NAME.password PASSWORD\n'
        '\n'
        '8. Copy Uploaded Torrents To Directory (Optional)\n'
        '\n'
        f'   $ {__project_name__} set trackers.{tracker_cls.name}.copy_to /path/to/directory'
    )

def test_Howto_upload():
    tracker_cls = type(make_MockTracker())
    howto = base._howto.Howto(tracker_cls=tracker_cls)
    howto._section = 123
    assert howto.upload == (
        '123. Upload\n'
        '\n'
        f'   $ {__project_name__} submit {tracker_cls.name} --help\n'
        f'   $ {__project_name__} submit {tracker_cls.name} /path/to/content'
    )
